import json
import re
import time

from platypush.backend import Backend
from platypush.message.event.music import MusicPlayEvent, MusicPauseEvent, \
    MusicStopEvent, NewPlayingTrackEvent, PlaylistChangeEvent, VolumeChangeEvent, \
    PlaybackConsumeModeChangeEvent, PlaybackSingleModeChangeEvent, \
    PlaybackRepeatModeChangeEvent, PlaybackRandomModeChangeEvent, \
    MuteChangeEvent, SeekChangeEvent


# noinspection PyUnusedLocal
class MusicMopidyBackend(Backend):
    """
    This backend listens for events on a Mopidy music server streaming port.
    Since this backend leverages the Mopidy websocket interface it is only
    compatible with Mopidy and not with other MPD servers. Please use the
    :class:`platypush.backend.music.mpd.MusicMpdBackend` for a similar polling
    solution if you're not running Mopidy or your instance has the websocket
    interface or web port disabled.

    Triggers:

        * :class:`platypush.message.event.music.MusicPlayEvent` if the playback state changed to play
        * :class:`platypush.message.event.music.MusicPauseEvent` if the playback state changed to pause
        * :class:`platypush.message.event.music.MusicStopEvent` if the playback state changed to stop
        * :class:`platypush.message.event.music.NewPlayingTrackEvent` if a new track is being played
        * :class:`platypush.message.event.music.PlaylistChangeEvent` if the main playlist has changed
        * :class:`platypush.message.event.music.VolumeChangeEvent` if the main volume has changed
        * :class:`platypush.message.event.music.MuteChangeEvent` if the mute status has changed
        * :class:`platypush.message.event.music.SeekChangeEvent` if a track seek event occurs

    Requires:
        * **websocket-client** (``pip install websocket-client``)
        * Mopidy installed and the HTTP service enabled
    """

    def __init__(self, host='localhost', port=6680, **kwargs):
        super().__init__(**kwargs)

        self.host = host
        self.port = int(port)
        self.url = 'ws://{}:{}/mopidy/ws'.format(host, port)
        self._msg_id = 0
        self._ws = None
        self._latest_status = {}
        self._connected = False

        try:
            self._latest_status = self._get_tracklist_status()
        except Exception as e:
            self.logger.warning('Unable to get mopidy status: {}'.format(str(e)))

    @staticmethod
    def _parse_track(track, pos=None):
        if not track:
            return {}

        conv_track = track.get('track', {}).copy()
        conv_track['id'] = track.get('tlid')
        conv_track['file'] = conv_track['uri']
        del conv_track['uri']

        if 'artists' in conv_track:
            conv_track['artist'] = conv_track['artists'][0].get('name')
            del conv_track['artists']

        if 'name' in conv_track:
            conv_track['title'] = conv_track['name']
            del conv_track['name']

        if 'album' in conv_track:
            conv_track['album'] = conv_track['album']['name']

        if 'length' in conv_track:
            conv_track['time'] = conv_track['length']/1000 \
                if conv_track['length'] else conv_track['length']
            del conv_track['length']

        if pos is not None:
            conv_track['pos'] = pos

        if '__model__' in conv_track:
            del conv_track['__model__']

        return conv_track

    def _communicate(self, msg):
        import websocket

        if isinstance(msg, str):
            msg = json.loads(msg)

        self._msg_id += 1
        msg['jsonrpc'] = '2.0'
        msg['id'] = self._msg_id
        msg = json.dumps(msg)

        ws = websocket.create_connection(self.url)
        ws.send(msg)
        response = json.loads(ws.recv()).get('result')
        ws.close()
        return response

    def _get_tracklist_status(self):
        return {
            'repeat': self._communicate({
                'method': 'core.tracklist.get_repeat'}),
            'random': self._communicate({
                'method': 'core.tracklist.get_random'}),
            'single': self._communicate({
                'method': 'core.tracklist.get_single'}),
            'consume': self._communicate({
                'method': 'core.tracklist.get_consume'}),
        }

    def _on_msg(self):
        def hndl(ws, msg):
            msg = json.loads(msg)
            event = msg.get('event')
            if not event:
                return

            status = {}
            track = msg.get('tl_track', {})

            if event == 'track_playback_paused':
                status['state'] = 'pause'
                track = self._parse_track(track)
                if not track:
                    return
                self.bus.post(MusicPauseEvent(status=status, track=track))
            elif event == 'track_playback_resumed':
                status['state'] = 'play'
                track = self._parse_track(track)
                if not track:
                    return
                self.bus.post(MusicPlayEvent(status=status, track=track))
            elif event == 'track_playback_ended' or (
                    event == 'playback_state_changed'
                    and msg.get('new_state') == 'stopped'):
                status['state'] = 'stop'
                track = self._parse_track(track)
                self.bus.post(MusicStopEvent(status=status, track=track))
            elif event == 'track_playback_started':
                track = self._parse_track(track)
                if not track:
                    return

                status['state'] = 'play'
                status['position'] = 0.0
                status['time'] = track.get('time')
                self.bus.post(NewPlayingTrackEvent(status=status, track=track))
            elif event == 'stream_title_changed':
                m = re.match('^\s*(.+?)\s+-\s+(.*)\s*$', msg.get('title', ''))
                if not m:
                    return

                track['artist'] = m.group(1)
                track['title'] = m.group(2)
                status['state'] = 'play'
                status['position'] = 0.0
                self.bus.post(NewPlayingTrackEvent(status=status, track=track))
            elif event == 'volume_changed':
                status['volume'] = msg.get('volume')
                self.bus.post(VolumeChangeEvent(volume=status['volume'],
                                                status=status, track=track))
            elif event == 'mute_changed':
                status['mute'] = msg.get('mute')
                self.bus.post(MuteChangeEvent(mute=status['mute'],
                                              status=status, track=track))
            elif event == 'seeked':
                status['position'] = msg.get('time_position')/1000
                self.bus.post(SeekChangeEvent(position=status['position'],
                                              status=status, track=track))
            elif event == 'tracklist_changed':
                tracklist = [self._parse_track(t, pos=i)
                             for i, t in enumerate(self._communicate({
                                'method': 'core.tracklist.get_tl_tracks'}))]

                self.bus.post(PlaylistChangeEvent(changes=tracklist))
            elif event == 'options_changed':
                new_status = self._get_tracklist_status()
                if new_status['random'] != self._latest_status.get('random'):
                    self.bus.post(PlaybackRandomModeChangeEvent(state=new_status['random']))
                if new_status['repeat'] != self._latest_status['repeat']:
                    self.bus.post(PlaybackRepeatModeChangeEvent(state=new_status['repeat']))
                if new_status['single'] != self._latest_status['single']:
                    self.bus.post(PlaybackSingleModeChangeEvent(state=new_status['single']))
                if new_status['consume'] != self._latest_status['consume']:
                    self.bus.post(PlaybackConsumeModeChangeEvent(state=new_status['consume']))

                self._latest_status = new_status

        return hndl

    def _on_error(self):
        def hndl(ws, error):
            self.logger.warning('Mopidy websocket error: {}'.format(error))
        return hndl

    def _on_close(self):
        def hndl():
            self._ws = None
            self._connected = False
            self.logger.warning('Mopidy websocket connection closed')

            while not self._connected:
                self._connect()
                time.sleep(10)

        return hndl

    def _on_open(self):
        def hndl(ws):
            self._connected = True
            self.logger.info('Mopidy websocket connected')
        return hndl

    def _connect(self):
        import websocket

        if not self._ws:
            self._ws = websocket.WebSocketApp(self.url,
                                              on_message=self._on_msg(),
                                              on_error=self._on_error(),
                                              on_close=self._on_close())

    def run(self):
        super().run()
        self.logger.info('Started tracking Mopidy events backend on {}:{}'.
                         format(self.host, self.port))

        self._connect()
        self._ws.on_open = self._on_open()
        self._ws.run_forever()


# vim:sw=4:ts=4:et:



import enum
import requests
import threading

from platypush.plugins import Plugin, action


class InterruptionFilterPolicy(enum.Enum):
    ALLOW_ALL = 1
    PRIORITY_ONLY = 2
    BLOCK_ALL = 3
    ALARM_ONLY = 4


class MobileJoinPlugin(Plugin):
    """
    Control mobile devices and other devices linked to Join (https://joaoapps.com/join/api/).
    It requires you to have either the Join app installed on an Android device or the Join
    browser extension installed.

    The ``device`` parameter in the actions can be:

    - A device ID or name
    - A list or comma-separated list of device IDs/names
    - A group name or a list of group names

    Supported groups:

        - group.all
        - group.android
        - group.windows10
        - group.phone
        - group.tablet
        - group.pc

    """

    _base_url = 'https://joinjoaomgcd.appspot.com/_ah/api'
    _push_url = '{}/messaging/v1/sendPush'.format(_base_url)
    _devices_url = '{}/registration/v1/listDevices'.format(_base_url)
    _groups = {
        'group.all',
        'group.android',
        'group.windows10',
        'group.phone',
        'group.tablet',
        'group.pc',
    }

    def __init__(self, api_key: str, **kwargs):
        """
        :param api_key: Join API key. Get your at https://joinjoaomgcd.appspot.com/.
        """
        super().__init__(**kwargs)

        self._api_key = api_key
        self._devices = {}
        self._devices_by_id = {}
        self._devices_by_name = {}
        self._devices_lock = threading.RLock()

    def _send_request(self, url, device=None, params: dict = None, **kwargs):
        if not params:
            params = {}

        params['apikey'] = self._api_key
        if device:
            params['deviceIds'] = self._get_device_ids(device)

        response = requests.get(url, params=params, **kwargs)
        if not response.ok:
            response.raise_for_status()

        response = response.json()
        if response.get('userAuthError'):
            raise PermissionError('Invalid api_key provided')

        del response['userAuthError']
        return response

    def _init_devices(self):
        with self._devices_lock:
            if not self._devices:
                self._devices = self.get_devices().output
                self._devices_by_id = {dev['id']: dev for dev in self._devices}
                self._devices_by_name = {dev['name']: dev for dev in self._devices}

    def _get_device_ids(self, device):
        devices = [dev.strip() for dev in device.split(',')] \
            if isinstance(device, str) else device
        assert isinstance(devices, list)

        has_unknown_devices = True
        cache_refreshed = False
        device_ids = []

        while has_unknown_devices:
            has_unknown_devices = False

            for dev in devices:
                if dev in self._devices_by_id:
                    device_ids.append(self._devices_by_id[dev]['id'])
                elif dev in self._devices_by_name:
                    device_ids.append(self._devices_by_name[dev]['id'])
                elif dev in self._groups:
                    device_ids.append(dev)
                else:
                    has_unknown_devices = True

            if has_unknown_devices:
                if cache_refreshed:
                    raise KeyError('No such device: {}'.format(device))

                self._init_devices()
                cache_refreshed = True

        return device_ids

    @action
    def push(self, device, text=None, url=None, file=None):
        """
        Push a URL or file to one or more devices

        :param device: Device ID or name, or list of device IDs/names, or group name(s)
        :param text: Optional description text for the URL or file
        :param url: URL to be pushed
        :param file: A publicly accessible URL of a file. You can also send the url of a file on your personal
            Google Drive
        """

        assert url or file
        params = {}

        if text:
            params['text'] = text
        if url:
            params['url'] = url
        if file:
            params['file'] = file

        return self._send_request(self._push_url, device=device, params=params)

    @action
    def set_clipboard(self, device, text):
        """
        Write to the clipboard of a device

        :param device: Device ID or name, or list of device IDs/names, or group name(s)
        :param text: Text to be set
        """
        return self._send_request(self._push_url, device=device, params={'clipboard': text})

    @action
    def send_sms(self, device, text: str, number: str = None, contact_name: str = None):
        """
        Send an sms through a mobile device connected to Join

        :param device: Device ID or name, or list of device IDs/names, or group name(s)
        :param text: Text to be sent
        :param number: Phone number
        :param contact_name: Alternatively to the phone number, you can specify a contact name
        """

        assert (number or contact_name) and not (number and contact_name)
        params = {'smstext': text}

        if number:
            params['smsnumber'] = number
        else:
            params['smscontactname'] = contact_name

        return self._send_request(self._push_url, device=device, params=params)

    @action
    def send_mms(self, device, file: str = None, text: str = None, subject: str = '',
                 number: str = None, contact_name: str = None, urgent: bool = False):
        """
        Send an MMS through a mobile device connected to Join

        :param device: Device ID or name, or list of device IDs/names, or group name(s)
        :param text: Text to be sent
        :param number: Phone number
        :param contact_name: Alternatively to the phone number, you can specify a contact name
        :param file: File attached to the message. Must be a local (to the phone) file or a publicly accessible URL
        :param subject: MMS subject
        :param urgent: Set to True if this is an urgent MMS. This will make the sent message be an MMS instead of an SMS
        """

        assert (number or contact_name) and not (number and contact_name)
        params = {
            'mmssubject': subject,
            'mmsurgent': int(urgent),
        }

        if number:
            params['smsnumber'] = number
        else:
            params['smscontactname'] = contact_name
        if file:
            params['mmsfile'] = file
        if text:
            params['smstext'] = text

        return self._send_request(self._push_url, device=device, params=params)

    @action
    def call_number(self, device, number: str):
        """
        Call a phone number through a mobile device connected to Join

        :param device: Device ID or name, or list of device IDs/names, or group name(s)
        :param number: Phone number
        """
        return self._send_request(self._push_url, device=device, params={'callnumber': number})

    @action
    def set_wallpaper(self, device, wallpaper: str):
        """
        Set the wallpaper on a device connected to Join

        :param device: Device ID or name, or list of device IDs/names, or group name(s)
        :param wallpaper: A publicly accessible URL of an image file. Will set the wallpaper on the receiving device
        """
        return self._send_request(self._push_url, device=device, params={'wallpaper': wallpaper})

    @action
    def set_lock_wallpaper(self, device, wallpaper: str):
        """
        Set the lock wallpaper on a device connected to Join

        :param device: Device ID or name, or list of device IDs/names, or group name(s)
        :param wallpaper: A publicly accessible URL of an image file. Will set the lockscreen wallpaper on the receiving
            device if the device has Android 7 or above
        """
        return self._send_request(self._push_url, device=device, params={'wallpaper': wallpaper})

    @action
    def find(self, device):
        """
        Make a device ring loudly
        :param device: Device ID or name, or list of device IDs/names, or group name(s)
        """
        return self._send_request(self._push_url, device=device, params={'find': True})

    @action
    def set_media_volume(self, device, volume: float):
        """
        Set media volume

        :param device: Device ID or name, or list of device IDs/names, or group name(s)
        :param volume: Volume
        """
        return self._send_request(self._push_url, device=device, params={'mediaVolume': volume})

    @action
    def set_ring_volume(self, device, volume: float):
        """
        Set ring volume

        :param device: Device ID or name, or list of device IDs/names, or group name(s)
        :param volume: Volume
        """
        return self._send_request(self._push_url, device=device, params={'ringVolume': volume})

    @action
    def set_alarm_volume(self, device, volume: float):
        """
        Set alarm volume

        :param device: Device ID or name, or list of device IDs/names, or group name(s)
        :param volume: Volume
        """
        return self._send_request(self._push_url, device=device, params={'alarmVolume': volume})

    @action
    def set_interruption_filter(self, device, policy: str):
        """
        Set interruption filter on one or more devices

        :param device: Device ID or name, or list of device IDs/names, or group name(s)
        :param policy: Possible values:

            - 'allow_all': Allow all notifications
            - 'priority_only': Allow only priority notifications and calls
            - 'alarm_only': Allow only alarm-related interruptions
            - 'block_all': Do not allow any interruptions

        """

        try:
            policy = getattr(InterruptionFilterPolicy, policy.upper())
        except AttributeError:
            raise AttributeError('Invalid policy: {}. Supported values: {}'.format(
                policy, [i.name for i in InterruptionFilterPolicy]
            ))

        return self._send_request(self._push_url, device=device, params={'interruptionFilter': policy})

    @action
    def say(self, device, text: str, language: str = None):
        """
        Say some text through a device's TTS engine

        :param device: Device ID or name, or list of device IDs/names, or group name(s)
        :param text: Text to say
        :param language: Language code
        """
        params = {'say': text}
        if language:
            params['language'] = language

        return self._send_request(self._push_url, device=device, params=params)

    @action
    def launch_app(self, device, name: str = None, package: str = None):
        """
        Launch an app on a device

        :param device: Device ID or name, or list of device IDs/names, or group name(s)
        :param name: Application name
        :param package: Alternatively to the application name, you can also specify the application
            package name. You can check the package name for an app by going to its Google Play page and checking
            the end of the URL. Example: for YouTube this is the URL
            (https://play.google.com/store/apps/details?id=com.google.android.youtube) and this is the package name
            (com.google.android.youtube)
        """

        assert (name or package) and not (name and package)

        params = {}
        if name:
            params['app'] = name
        else:
            params['appPackage'] = package

        return self._send_request(self._push_url, device=device, params=params)

    @action
    def send_notification(self, device, title: str = None, text: str = None, url: str = None, file: str = None,
                          icon: str = None, small_icon: str = None, priority: int = 2, vibration_pattern=None,
                          dismiss_on_touch: bool = False, image: str = None, group: str = None,
                          sound: str = None, actions=None):
        """
        Send a notification to a device

        :param device: Device ID or name, or list of device IDs/names, or group name(s)
        :param title: Notification title
        :param text: Notification text
        :param url: URL to be opened on touch
        :param file: A publicly accessible URL of a file that will be opened or downloaded on touch. You can also
            send the url of a file on your personal Google Drive.
        :param icon: If a notification is created on the receiving device and this is set, then it'll be used as
            the notification’s icon. If this image has transparency, it’ll also be used as the status bar icon is
            smallicon is not set. It’ll also be used to tint the notification with its dominating color
        :param small_icon: If a notification is created on the receiving device and this is set, then it’ll be used as
            the notification’s status bar icon
        :param priority: Control how your notification is displayed: lower priority notifications are usually
            displayed lower in the notification list. Values from -2 (lowest priority) to 2 (highest priority).
            Default is 2.
        :param vibration_pattern: If the notification is received on an Android device, the vibration pattern in this
            field will change the way the device vibrates with it. You can easily create a pattern by going to the
            `AutoRemote notification page <http://autoremotejoaomgcd.appspot.com/AutoRemoteNotification.html>`_
            and generate the pattern in the Vibration Pattern field
        :type vibration_pattern: str (comma-separated float values) or list[float]
        :param dismiss_on_touch: If set the notification will be dismissed when touched (default: False)
        :param image: Publicly available URL for an image to show up in the notification
        :param group: Unique ID to group your notifications with
        :param sound: Publicly available URL for a sound to play with the notification
        :param actions: Set notification buttons with customized behaviour. This parameter is a list of Join actions
            configured on the target device that will be mapped to notification input elements.
            More info `on the Joaoapps notifications page <https://joaoapps.com/join/actions/#notifications>`_
        """

        params = {'dismissOnTouch': dismiss_on_touch}

        if title:
            params['title'] = title
        if text:
            params['text'] = text
        if url:
            params['url'] = url
        if file:
            params['file'] = file
        if icon:
            params['icon'] = icon
        if small_icon:
            params['smallIcon'] = small_icon
        if priority:
            params['priority'] = priority
        if vibration_pattern:
            params['vibrationPattern'] = [i for i in vibration_pattern.split(',') if len(i)] \
                if isinstance(vibration_pattern, str) else vibration_pattern
            assert isinstance(params['vibrationPattern'], list)
        if image:
            params['image'] = image
        if group:
            params['group'] = group
        if sound:
            params['sound'] = sound
        if actions:
            actions = [a.strip() for a in actions.split('|||' if '|||' in actions else ',') if len(a.strip())] \
                if isinstance(actions, str) else actions
            assert isinstance(actions, list) and len(actions) > 0
            params['actions'] = '|||'.join(actions)

        return self._send_request(self._push_url, device=device, params=params)

    @action
    def get_devices(self):
        """
        :return: List of connected devices, each containing 'id', 'name', 'user' and 'has_tasker' attributes
        """

        return [
            {
                'id': dev['deviceId'],
                'name': dev['deviceName'],
                'user': dev['userAccount'],
                'has_tasker': dev['hasTasker'],
            }
            for dev in self._send_request(self._devices_url)['records']
        ]


# vim:sw=4:ts=4:et:



import datetime
import hashlib
import random

import bcrypt

from sqlalchemy import Column, Integer, String, DateTime, ForeignKey
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.ext.declarative import declarative_base

from platypush.context import get_plugin

Base = declarative_base()


class UserManager:
    """
    Main class for managing platform users
    """

    # noinspection PyProtectedMember
    def __init__(self):
        db_plugin = get_plugin('db')
        if not db_plugin:
            raise ModuleNotFoundError('Please enable/configure the db plugin for multi-user support')

        self._engine = db_plugin._get_engine()

    def get_user(self, username):
        session = self._get_db_session()
        user = self._get_user(session, username)
        if not user:
            return None

        # Hide password
        user.password = None
        return user

    def get_user_count(self):
        session = self._get_db_session()
        return session.query(User).count()

    def get_users(self):
        session = self._get_db_session()
        return session.query(User)

    def create_user(self, username, password, **kwargs):
        session = self._get_db_session()
        if not username:
            raise ValueError('Invalid or empty username')
        if not password:
            raise ValueError('Please provide a password for the user')

        user = self._get_user(session, username)
        if user:
            raise NameError('The user {} already exists'.format(username))

        record = User(username=username, password=self._encrypt_password(password),
                      created_at=datetime.datetime.utcnow(), **kwargs)

        session.add(record)
        session.commit()
        user = self._get_user(session, username)

        # Hide password
        user.password = None
        return user

    def update_password(self, username, old_password, new_password):
        session = self._get_db_session()
        if not self._authenticate_user(session, username, old_password):
            return False

        user = self._get_user(session, username)
        user.password = self._encrypt_password(new_password)
        session.commit()
        return True

    def authenticate_user(self, username, password):
        session = self._get_db_session()
        return self._authenticate_user(session, username, password)

    def authenticate_user_session(self, session_token):
        session = self._get_db_session()
        user_session = session.query(UserSession).filter_by(session_token=session_token).first()

        if not user_session or (
                user_session.expires_at and user_session.expires_at < datetime.datetime.utcnow()):
            return None, None

        user = session.query(User).filter_by(user_id=user_session.user_id).first()

        # Hide password
        user.password = None
        return user, session

    def delete_user(self, username):
        session = self._get_db_session()
        user = self._get_user(session, username)
        if not user:
            raise NameError('No such user: {}'.format(username))

        user_sessions = session.query(UserSession).filter_by(user_id=user.user_id).all()
        for user_session in user_sessions:
            session.delete(user_session)

        session.delete(user)
        session.commit()
        return True

    def delete_user_session(self, session_token):
        session = self._get_db_session()
        user_session = session.query(UserSession).filter_by(session_token=session_token).first()

        if not user_session:
            return False

        session.delete(user_session)
        session.commit()
        return True

    def create_user_session(self, username, password, expires_at=None):
        session = self._get_db_session()
        if not self._authenticate_user(session, username, password):
            return None

        if expires_at:
            if isinstance(expires_at, int) or isinstance(expires_at, float):
                expires_at = datetime.datetime.fromtimestamp(expires_at)
            elif isinstance(expires_at, str):
                expires_at = datetime.datetime.fromisoformat(expires_at)

        user = self._get_user(session, username)
        user_session = UserSession(user_id=user.user_id, session_token=self._generate_token(),
                                   csrf_token=self._generate_token(), created_at=datetime.datetime.utcnow(),
                                   expires_at=expires_at)

        session.add(user_session)
        session.commit()
        return user_session

    @staticmethod
    def _get_user(session, username):
        return session.query(User).filter_by(username=username).first()

    @staticmethod
    def _encrypt_password(pwd):
        return bcrypt.hashpw(pwd.encode(), bcrypt.gensalt(12))

    @staticmethod
    def _check_password(pwd, hashed_pwd):
        return bcrypt.checkpw(pwd.encode(), hashed_pwd)

    @staticmethod
    def _generate_token():
        rand = bytes(random.randint(0, 255) for _ in range(0, 255))
        return hashlib.sha256(rand).hexdigest()

    def _get_db_session(self):
        Base.metadata.create_all(self._engine)
        session = scoped_session(sessionmaker())
        session.configure(bind=self._engine)
        return session()

    def _authenticate_user(self, session, username, password):
        user = self._get_user(session, username)
        if not user:
            return False

        return self._check_password(password, user.password)


class User(Base):
    """ Models the User table """

    __tablename__ = 'user'
    __table_args__ = ({'sqlite_autoincrement': True})

    user_id = Column(Integer, primary_key=True)
    username = Column(String, unique=True, nullable=False)
    password = Column(String)
    created_at = Column(DateTime)


class UserSession(Base):
    """ Models the UserSession table """

    __tablename__ = 'user_session'
    __table_args__ = ({'sqlite_autoincrement': True})

    session_id = Column(Integer, primary_key=True)
    session_token = Column(String, unique=True, nullable=False)
    csrf_token = Column(String, unique=True)
    user_id = Column(Integer, ForeignKey('user.user_id'), nullable=False)
    created_at = Column(DateTime)
    expires_at = Column(DateTime)


# vim:sw=4:ts=4:et:



from typing import Optional, Dict, Any, List, Union

from platypush.backend.alarm import AlarmBackend
from platypush.context import get_backend
from platypush.plugins import Plugin, action


class AlarmPlugin(Plugin):
    """
    Alarm/timer plugin.

    Requires:

        - The :class:`platypush.backend.alarm.AlarmBackend` backend configured and enabled.

    """

    @staticmethod
    def _get_backend() -> AlarmBackend:
        return get_backend('alarm')

    @action
    def add(self, when: str, actions: Optional[list] = None, name: Optional[str] = None,
            audio_file: Optional[str] = None, audio_volume: Optional[Union[int, float]] = None,
            enabled: bool = True) -> str:
        """
        Add a new alarm. NOTE: alarms that aren't configured in the :class:`platypush.backend.alarm.AlarmBackend`
        will only run in the current session. If you want an alarm to be permanently stored, you should configure
        it in the alarm backend configuration. You may want to add an alarm dynamically if it's a one-time alarm instead

        :param when: When the alarm should be executed. It can be either a cron expression (for recurrent alarms), or
            a datetime string in ISO format (for one-shot alarms/timers), or an integer representing the number of
            seconds before the alarm goes on (e.g. 300 for 5 minutes).
        :param actions: List of actions to be executed.
        :param name: Alarm name.
        :param audio_file: Path of the audio file to be played.
        :param audio_volume: Volume of the audio.
        :param enabled: Whether the new alarm should be enabled (default: True).
        :return: The alarm name.
        """
        alarm = self._get_backend().add_alarm(when=when, audio_file=audio_file, actions=actions or [],
                                              name=name, enabled=enabled, audio_volume=audio_volume)
        return alarm.name

    @action
    def enable(self, name: str):
        """
        Enable an alarm.

        :param name: Alarm name.
        """
        self._get_backend().enable_alarm(name)

    @action
    def disable(self, name: str):
        """
        Disable an alarm. This will prevent the alarm from executing until re-enabled or until the application
        is restarted.

        :param name: Alarm name.
        """
        self._get_backend().disable_alarm(name)

    @action
    def dismiss(self):
        """
        Dismiss the alarm that is currently running.
        """
        self._get_backend().dismiss_alarm()

    @action
    def snooze(self, interval: Optional[float] = 300.0):
        """
        Snooze the alarm that is currently running for the specified number of seconds.
        The alarm will stop and resume again later.

        :param interval: Snooze seconds before playing the alarm again (default: 300).
        """
        self._get_backend().snooze_alarm(interval=interval)

    @action
    def get_alarms(self) -> List[Dict[str, Any]]:
        """
        Get the list of configured alarms.

        :return: List of the alarms, sorted by next scheduled run.
        """
        return [alarm.to_dict() for alarm in self._get_backend().get_alarms()]


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class KafkaMessageEvent(Event):
    """
    Kafka message event object. Fired when :mod:`platypush.backend.kafka` receives
    a new event.
    """

    def __init__(self, msg, *args, **kwargs):
        """
        :param msg: Received message
        :type msg: str or bytes stream
        """

        super().__init__(msg=msg, *args, **kwargs)


# vim:sw=4:ts=4:et:




from platypush.plugins import Plugin, action

class LightPlugin(Plugin):
    """
    Abstract plugin to interface your logic with lights/bulbs.
    """

    @action
    def on(self):
        """ Turn the light on """
        raise NotImplementedError()

    @action
    def off(self):
        """ Turn the light off """
        raise NotImplementedError()

    @action
    def toggle(self):
        """ Toggle the light status (on/off) """
        raise NotImplementedError()

    @action
    def status(self):
        """ Get the light status """
        raise NotImplementedError()


# vim:sw=4:ts=4:et:




import requests

from platypush.plugins import Plugin, action


class IftttPlugin(Plugin):
    """
    This plugin allows you to interact with the IFTTT maker API
    <https://ifttt.com/maker_webhooks> to programmatically trigger your own
    IFTTT hooks from Platypush - e.g. send a tweet or a Facebook post, create a
    Todoist item or a Trello task, trigger events on your mobile device, or run
    any action not natively supported by Platypush but available on your IFTTT
    configuration.

    Requires:

        * **requests** (``pip install requests``)

    An example::

        # Trigger an IFTTT event named "at_home"
        {
            "type": "request",
            "action": "ifttt.trigger_event",
            "args": {
                "event_name": "at_home"
            }
        }
    """

    _base_url = 'https://maker.ifttt.com/trigger/{event_name}/with/key/{ifttt_key}'

    def __init__(self, ifttt_key, **kwargs):
        """
        :param ifttt_key: Your IFTTT Maker API key. Log in to IFTTT and get your key from
            `here <https://ifttt.com/maker_webhooks>`_. Once you've got your key, you can start creating IFTTT rules
            using the Webhooks channel.
        :type ifttt_key: str
        """

        super().__init__(**kwargs)
        self.ifttt_key = ifttt_key

    @action
    def trigger_event(self, event_name, values=None):
        """
        Send an event to your IFTTT account

        :param event_name: Name of the event
        :type event_name: str

        :param values: Optional list of values to be passed to the event. By convention IFTTT names the values as
            ``value1,value2,...``.
        :type values: list
        """

        url = self._base_url.format(event_name=event_name, ifttt_key=self.ifttt_key)
        if not values:
            values = []

        response = requests.post(url, json={'value{}'.format(i + 1): v
                                            for (i, v) in enumerate(values)})

        if not response.ok:
            raise RuntimeError("IFTTT event '{}' error: {}: {}".format(
                event_name, response.status_code, response.reason))


# vim:sw=4:ts=4:et:



import os
import subprocess
import tempfile
from typing import Optional, List

from platypush.plugins import action
from platypush.plugins.tts import TtsPlugin


class TtsGooglePlugin(TtsPlugin):
    """
    Advanced text-to-speech engine that leverages the Google Cloud TTS API.
    See https://cloud.google.com/text-to-speech/docs/quickstart-client-libraries#client-libraries-install-python
    for how to enable the API on your account and get your credentials.

    Requires:

        * **google-cloud-texttospeech** - ``pip install google-cloud-texttospeech``
        * **mplayer** - see your distribution docs on how to install the mplayer package
    """

    def __init__(self,
                 language: str ='en-US',
                 voice: Optional[str] = None,
                 gender: str = 'FEMALE',
                 credentials_file: str = '~/.credentials/platypush/google/platypush-tts.json',
                 player_args: Optional[List[str]] = None):
        """
        :param language: Language code, see https://cloud.google.com/text-to-speech/docs/basics for supported languages
        :param voice: Voice type, see https://cloud.google.com/text-to-speech/docs/basics for supported voices
        :param gender: Voice gender (MALE, FEMALE or NEUTRAL)
        :param credentials_file: Where your GCloud credentials for TTS are stored, see https://cloud.google.com/text-to-speech/docs/basics
        :param player_args: Extra options to be passed to the audio player (default: ``mplayer``).
        """
        from google.cloud import texttospeech
        super().__init__()

        self.language = language
        self.voice = voice
        self.player_args = player_args or []

        self.language = self._parse_language(language)
        self.voice = self._parse_voice(self.language, voice)
        self.gender = getattr(texttospeech.enums.SsmlVoiceGender, gender.upper())
        os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = os.path.expanduser(credentials_file)

    def _parse_language(self, language):
        if language is None:
            language = self.language or 'en-US'

        if len(language) == 2:
            language = language.lower()
            if language == 'en':
                language = 'en-US'
            else:
                language += '-' + language.upper()

        return language

    @staticmethod
    def _parse_voice(language, voice):
        if voice is not None:
            return voice

        if language == 'en-US':
            return language + '-Wavenet-C'
        return language + '-Wavenet-A'

    @action
    def say(self,
            text: str,
            language: Optional[str] = None,
            voice: Optional[str] = None,
            gender: Optional[str] = None,
            player_args: Optional[List[str]] = None):
        """
        Say a phrase.

        :param text: Text to say.
        :param language: Language code override.
        :param voice: Voice type override.
        :param gender: Gender override.
        :param player_args: Player args override.
        """

        from google.cloud import texttospeech
        client = texttospeech.TextToSpeechClient()
        synthesis_input = texttospeech.types.SynthesisInput(text=text)

        language = self._parse_language(language)
        voice = self._parse_voice(language, voice)

        if gender is None:
            gender = self.gender
        else:
            gender = getattr(texttospeech.enums.SsmlVoiceGender, gender.upper())

        player_args = player_args or self.player_args
        voice = texttospeech.types.VoiceSelectionParams(
            language_code=language, ssml_gender=gender,
            name=voice)

        audio_config = texttospeech.types.AudioConfig(
            audio_encoding=texttospeech.enums.AudioEncoding.MP3)

        response = client.synthesize_speech(synthesis_input, voice, audio_config)

        with tempfile.NamedTemporaryFile() as f:
            f.write(response.audio_content)
            cmd = ['mplayer -ao alsa -really-quiet -noconsolecontrols {} "{}"'.format(
                ' '.join(player_args), f.name)]

            try:
                return subprocess.check_output(
                    cmd, stderr=subprocess.STDOUT, shell=True).decode('utf-8')
            except subprocess.CalledProcessError as e:
                raise RuntimeError(e.output.decode('utf-8'))


# vim:sw=4:ts=4:et:



"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
.. license: MIT
"""

import logging
import threading
import time

from threading import Thread
from typing import Optional

from platypush.bus import Bus
from platypush.config import Config
from platypush.context import get_backend
from platypush.utils import set_timeout, clear_timeout, \
    get_redis_queue_name_by_message, set_thread_name

from platypush.event import EventGenerator
from platypush.message import Message
from platypush.message.event import Event, StopEvent
from platypush.message.request import Request
from platypush.message.response import Response


class Backend(Thread, EventGenerator):
    """
    Parent class for backends.

    A backend is basically a thread that checks for new events on some channel
    (e.g. a network socket, a queue, some new entries on an API endpoint or an
    RSS feed, a voice command through an assistant, a new measure from a sensor
    etc.) and propagates event messages to the main application bus whenever a
    new event happens. You can then build whichever type of custom logic you
    want on such events.
    """

    _default_response_timeout = 5

    # Loop function, can be implemented by derived classes
    loop = None

    def __init__(self, bus: Optional[Bus] = None, poll_seconds: Optional[float] = None, **kwargs):
        """
        :param bus: Reference to the bus object to be used in the backend
        :param poll_seconds: If the backend implements a ``loop`` method, this parameter expresses how often the
            loop should run in seconds.
        :param kwargs: Key-value configuration for the backend
        """

        self._thread_name = self.__class__.__name__
        EventGenerator.__init__(self)
        Thread.__init__(self, name=self._thread_name)

        # If no bus is specified, create an internal queue where
        # the received messages will be pushed
        self.bus = bus or Bus()
        self.poll_seconds = float(poll_seconds) if poll_seconds else None
        self.device_id = Config.get('device_id')
        self.thread_id = None
        self._should_stop = False
        self._stop_event = threading.Event()
        self._kwargs = kwargs
        self.logger = logging.getLogger(self.__class__.__name__)

        # Internal-only, we set the request context on a backend if that
        # backend is intended to react for a response to a specific request
        self._request_context = kwargs['_req_ctx'] if '_req_ctx' in kwargs \
            else None

        if 'logging' in kwargs:
            self.logger.setLevel(getattr(logging, kwargs.get('logging').upper()))

        Thread.__init__(self)

    def on_message(self, msg):
        """
        Callback when a message is received on the backend.
        It parses and posts the message on the main bus.
        It should be called by the derived classes whenever
        a new message should be processed.

        :param msg: Received message.  It can be either a key-value dictionary, a platypush.message.Message object,
            or a string/byte UTF-8 encoded string
        """

        msg = Message.build(msg)

        if not getattr(msg, 'target') or msg.target != self.device_id:
            return  # Not for me

        self.logger.debug('Message received on the {} backend: {}'.format(
            self.__class__.__name__, msg))

        if self._is_expected_response(msg):
            # Expected response, trigger the response handler
            clear_timeout()
            self._request_context['on_response'](msg)
            self.stop()
            return

        if isinstance(msg, StopEvent) and msg.targets_me():
            self.logger.info('Received STOP event on {}'.format(self.__class__.__name__))
            self._should_stop = True
        else:
            msg.backend = self   # Augment message to be able to process responses
            self.bus.post(msg)

    def _is_expected_response(self, msg):
        """ Internal only - returns true if we are expecting for a response
            and msg is that response """

        return self._request_context \
            and isinstance(msg, Response) \
            and msg.id == self._request_context['request'].id

    def _get_backend_config(self):
        config_name = 'backend.' + self.__class__.__name__.split('Backend')[0].lower()
        return Config.get(config_name)

    def _setup_response_handler(self, request, on_response, response_timeout):
        def _timeout_hndl():
            raise RuntimeError('Timed out while waiting for a response from {}'.
                               format(request.target))

        req_ctx = {
            'request': request,
            'on_response': on_response,
            'response_timeout': response_timeout,
        }

        resp_backend = self.__class__(bus=self.bus, _req_ctx=req_ctx,
                                      **self._get_backend_config(), **self._kwargs)

        # Set the response timeout
        if response_timeout:
            set_timeout(seconds=response_timeout, on_timeout=_timeout_hndl)

        resp_backend.start()

    def send_event(self, event, **kwargs):
        """
        Send an event message on the backend.

        :param event: Event to send. It can be a dict, a string/bytes UTF-8 JSON, or a platypush.message.event.Event
            object.
        """

        event = Event.build(event)
        assert isinstance(event, Event)

        event.origin = self.device_id
        if not hasattr(event, 'target'):
            event.target = self.device_id

        self.send_message(event, **kwargs)

    def send_request(self, request, on_response=None,
                     response_timeout=_default_response_timeout, **kwargs):
        """
        Send a request message on the backend.

        :param request: The request, either a dict, a string/bytes UTF-8 JSON, or a platypush.message.request.Request
            object.

        :param on_response: Optional callback that will be called when a response is received. If set, this method will
            synchronously wait for a response before exiting.
        :type on_response: function

        :param response_timeout: If on_response is set, the backend will raise an exception if the response isn't
            received within this number of seconds (default: None)
        :type response_timeout: float
        """

        request = Request.build(request)
        assert isinstance(request, Request)

        request.origin = self.device_id

        if on_response and response_timeout != 0:
            self._setup_response_handler(request, on_response, response_timeout)

        self.send_message(request, **kwargs)

    def send_response(self, response, request, **kwargs):
        """
        Send a response message on the backend.

        :param response: The response, either a dict, a string/bytes UTF-8 JSON, or a
            :class:`platypush.message.response.Response` object.
        :param request: Associated request, used to set the response parameters that will link them
        """

        assert isinstance(response, Response)
        assert isinstance(request, Request)

        self.send_message(response, **kwargs)

    def send_message(self, msg, queue_name=None, **kwargs):
        """
        Sends a platypush.message.Message to a node.
        To be implemented in the derived classes. By default, if the Redis
        backend is configured then it will try to deliver the message to
        other consumers through the configured Redis main queue.

        :param msg: The message to send
        :param queue_name: Send the message on a specific queue (default: the queue_name configured on the Redis
            backend)
        """

        try:
            redis = get_backend('redis')
            if not redis:
                raise KeyError()
        except KeyError:
            self.logger.warning("Backend {} does not implement send_message " +
                                "and the fallback Redis backend isn't configured")
            return

        redis.send_message(msg, queue_name=queue_name)

    def run(self):
        """ Starts the backend thread. To be implemented in the derived classes if the loop method isn't defined. """
        self.thread_id = threading.get_ident()
        set_thread_name(self._thread_name)
        if not callable(self.loop):
            return

        while not self.should_stop():
            try:
                with self:
                    has_error = False

                    while not self.should_stop() and not has_error:
                        try:
                            self.loop()
                        except Exception as e:
                            has_error = True
                            self.logger.error(str(e))
                            self.logger.exception(e)
                        finally:
                            if self.poll_seconds:
                                time.sleep(self.poll_seconds)
                            elif has_error:
                                time.sleep(5)
            except Exception as e:
                self.logger.error('{} initialization error: {}'.format(self.__class__.__name__, str(e)))
                self.logger.exception(e)
                time.sleep(self.poll_seconds or 5)

    def __enter__(self):
        """ Invoked when the backend is initialized, if the main logic is within a ``loop()`` function """
        self.logger.info('Initialized backend {}'.format(self.__class__.__name__))

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ Invoked when the backend is terminated, if the main logic is within a ``loop()`` function """
        self.on_stop()
        self.logger.info('Terminated backend {}'.format(self.__class__.__name__))

    def on_stop(self):
        """ Callback invoked when the process stops """
        pass

    def stop(self):
        """ Stops the backend thread by sending a STOP event on its bus """
        def _async_stop():
            evt = StopEvent(target=self.device_id, origin=self.device_id,
                            thread_id=self.thread_id)

            self.send_message(evt)
            self._stop_event.set()
            self.on_stop()

        Thread(target=_async_stop).start()

    def should_stop(self):
        return self._should_stop

    def wait_stop(self, timeout=None):
        self._stop_event.wait(timeout)

    def _get_redis(self):
        import redis

        redis_backend = get_backend('redis')
        if not redis_backend:
            self.logger.warning('Redis backend not configured - some ' +
                                'web server features may not be working properly')
            redis_args = {}
        else:
            redis_args = redis_backend.redis_args

        redis = redis.Redis(**redis_args)
        return redis

    def get_message_response(self, msg):
        try:
            redis = self._get_redis()
            response = redis.blpop(get_redis_queue_name_by_message(msg), timeout=60)
            if response and len(response) > 1:
                response = Message.build(response[1])
            else:
                response = None

            return response
        except Exception as e:
            self.logger.error('Error while processing response to {}: {}'.format(msg, str(e)))


# vim:sw=4:ts=4:et:



import base64
import json
import socket

from typing import Optional, Union

from platypush.plugins import Plugin, action


class UdpPlugin(Plugin):
    """
    Plugin for raw UDP communications.
    """

    @action
    def send(self, data: Union[bytes, str], host: str, port: int, binary: bool = False,
             timeout: Optional[float] = None, recv_response: bool = False, **recv_opts):
        """
        Send data over a UDP connection.

        :param data: Data to be sent, as bytes or string.
        :param host: Host IP/name.
        :param port: TCP port.
        :param binary: If set to True and ``data`` is a string then will be treated as base64-encoded binary input.
        :param timeout: Connection timeout in seconds (default: None).
        :param recv_response: If True then the action will wait for a response from the server before closing the
            connection. Note that ``recv_opts`` must be specified in this case - at least ``length``.
        """
        if isinstance(data, list) or isinstance(data, dict):
            data = json.dumps(data)
        if isinstance(data, str):
            data = data.encode()
            if binary:
                data = base64.decodebytes(data)

        sd = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        if timeout:
            sd.settimeout(timeout)

        sd.sendto(data, (host, port))
        if not recv_response:
            return

        recv_opts.update({
            'host': host,
            'port': port,
            'timeout': timeout,
            'binary': binary,
        })

        data = sd.recvfrom(**recv_opts)
        if binary:
            data = base64.encodebytes(data)
        data = data.decode()
        return data


# vim:sw=4:ts=4:et:



import threading
import time

from typing import Optional

from platypush.context import get_bus
from platypush.message.event.distance import DistanceSensorEvent
from platypush.plugins import action
from platypush.plugins.gpio import GpioPlugin
from platypush.plugins.gpio.sensor import GpioSensorPlugin


class GpioSensorDistancePlugin(GpioPlugin, GpioSensorPlugin):
    """
    You can use this plugin to interact with a distance sensor on your Raspberry
    Pi (tested with a HC-SR04 ultrasound sensor).

    Requires:

        * ``RPi.GPIO`` (``pip install RPi.GPIO``)

    Triggers:

        * :class:`platypush.message.event.distance.DistanceSensorEvent` when a new distance measurement is available

    """

    def __init__(self, trigger_pin: int, echo_pin: int, measurement_interval: float = 0.15,
                 timeout: float = 2.0, warmup_time: float = 2.0, *args, **kwargs):
        """
        :param trigger_pin: GPIO PIN where you connected your sensor trigger PIN (the one that triggers the
            sensor to perform a measurement).

        :param echo_pin: GPIO PIN where you connected your sensor echo PIN (the one that will listen for the
            signal to bounce back and therefore trigger the distance calculation).

        :param measurement_interval: When running in continuous mode (see
            :func:`platypush.plugins.gpio.sensor.distance.GpioSensorDistancePlugin.start_measurement`) this parameter
            indicates how long should be waited between two measurements (default: 0.15 seconds)
        :param timeout: The echo-wait will terminate and the plugin will return null if no echo has been
            received after this time (default: 1 second).

        :param warmup_time: Number of seconds that should be waited on plugin instantiation
            for the sensor to be ready (default: 2 seconds).
        """

        GpioPlugin.__init__(self, pins={'trigger': trigger_pin, 'echo': echo_pin, }, *args, **kwargs)

        self.trigger_pin = trigger_pin
        self.echo_pin = echo_pin
        self.measurement_interval = measurement_interval
        self.timeout = timeout
        self.warmup_time = warmup_time
        self._measurement_thread = None
        self._measurement_thread_lock = threading.RLock()
        self._measurement_thread_can_run = False
        self._init_board()

    def _init_board(self):
        import RPi.GPIO as GPIO

        with self._init_lock:
            if self._initialized:
                return

            GpioPlugin._init_board(self)
            self._initialized = False

            GPIO.setup(self.trigger_pin, GPIO.OUT)
            GPIO.setup(self.echo_pin, GPIO.IN)
            GPIO.output(self.trigger_pin, GPIO.LOW)

            self.logger.info('Waiting {} seconds for the sensor to be ready'.format(self.warmup_time))
            time.sleep(self.warmup_time)
            self.logger.info('Sensor ready')
            self._initialized = True

    def _get_data(self):
        import RPi.GPIO as GPIO

        pulse_start = pulse_on = time.time()

        self._init_board()
        GPIO.output(self.trigger_pin, GPIO.HIGH)
        time.sleep(0.00001)  # 1 us pulse to trigger echo measurement
        GPIO.output(self.trigger_pin, GPIO.LOW)

        while GPIO.input(self.echo_pin) == 0:
            pulse_on = time.time()
            if pulse_on - pulse_start > self.timeout:
                raise TimeoutError('Distance sensor echo timeout after {} seconds: 0'.
                                   format(self.timeout))

        pulse_start = pulse_on
        pulse_end = pulse_off = time.time()

        while GPIO.input(self.echo_pin) == 1:
            pulse_off = time.time()
            if pulse_off - pulse_end > self.timeout:
                raise TimeoutError('Distance sensor echo timeout after {} seconds: 1'.
                                   format(self.timeout))

        pulse_end = pulse_off
        pulse_duration = pulse_end - pulse_start

        # s = vt where v = 1/2 * avg speed of sound in mm/s
        return round(pulse_duration * 171500.0, 2)

    @action
    def get_measurement(self):
        """
        Extends :func:`.GpioSensorPlugin.get_measurement`

        :returns: Distance measurement as a scalar (in mm):
        """

        try:
            distance = self._get_data()
            bus = get_bus()
            bus.post(DistanceSensorEvent(distance=distance, unit='mm'))
            return distance
        except TimeoutError as e:
            self.logger.warning(str(e))
            return
        except Exception as e:
            self.close()
            raise e

    @action
    def close(self):
        return self.cleanup()

    def __enter__(self):
        self._init_board()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _get_measurement_thread(self, duration: float):
        def _thread():
            with self:
                self.logger.info('Started distance measurement thread')
                start_time = time.time()

                try:
                    while self._measurement_thread_can_run and (
                            not duration or time.time() - start_time <= duration):
                        self.get_measurement()
                        if self.measurement_interval:
                            time.sleep(self.measurement_interval)
                finally:
                    self._measurement_thread = None

        return _thread

    def _is_measurement_thread_running(self):
        with self._measurement_thread_lock:
            return self._measurement_thread is not None

    @action
    def start_measurement(self, duration: Optional[float] = None):
        """
        Start the measurement thread. It will trigger :class:`platypush.message.event.distance.DistanceSensorEvent`
        events when new measurements are available.

        :param duration: If set, then the thread will run for the specified amount of seconds (default: None)
        """
        with self._measurement_thread_lock:
            if self._is_measurement_thread_running():
                self.logger.warning('A measurement thread is already running')
                return

            thread_func = self._get_measurement_thread(duration=duration)
            self._measurement_thread = threading.Thread(target=thread_func)
            self._measurement_thread_can_run = True
            self._measurement_thread.start()

    @action
    def stop_measurement(self):
        """
        Stop the running measurement thread.
        """
        with self._measurement_thread_lock:
            if not self._is_measurement_thread_running():
                self.logger.warning('No measurement thread is running')
                return

            self._measurement_thread_can_run = False
            self.logger.info('Waiting for the measurement thread to end')

            if self._measurement_thread:
                self._measurement_thread.join(timeout=self.timeout)

            self.logger.info('Measurement thread terminated')


# vim:sw=4:ts=4:et:



from typing import Optional, List

from platypush.message.response import Response


class PrinterResponse(Response):
    def __init__(self,
                 *args,
                 name: str,
                 printer_type: int,
                 info: str,
                 uri: str,
                 state: int,
                 is_shared: bool,
                 state_message: Optional[str] = None,
                 state_reasons: Optional[List[str]] = None,
                 location: Optional[str] = None,
                 uri_supported: Optional[str] = None,
                 make_and_model: Optional[str] = None,
                 **kwargs):
        super().__init__(*args, output={
            'name': name,
            'printer_type': printer_type,
            'info': info,
            'uri': uri,
            'state': state,
            'is_shared': is_shared,
            'state_message': state_message,
            'state_reasons': state_reasons,
            'location': location,
            'uri_supported': uri_supported,
            'make_and_model': make_and_model,
        }, **kwargs)


class PrintersResponse(Response):
    def __init__(self,
                 *args,
                 printers: List[PrinterResponse],
                 **kwargs):
        super().__init__(*args, output={p.output['name']: p.output for p in printers}, **kwargs)


class PrinterJobAddedResponse(Response):
    def __init__(self,
                 *args,
                 printer: str,
                 job_id: int,
                 **kwargs):
        super().__init__(*args, output={
            'printer': printer,
            'job_id': job_id,
        }, **kwargs)


# vim:sw=4:ts=4:et:



from platypush.backend.sensor import SensorBackend
from platypush.context import get_plugin


class SensorBatteryBackend(SensorBackend):
    """
    This backend listens for battery full/connect/disconnect/below/above threshold events.
    The sensor events triggered by this backend will include any of the following fields:

        - ``battery_percent``
        - ``battery_power_plugged``

    Requires:

        - **psutil** (``pip install psutil``) for CPU load and stats.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_measurement(self):
        plugin = get_plugin('system')
        battery = plugin.sensors_battery().output

        return {
            'battery_percent': battery.get('percent'),
            'battery_power_plugged': bool(battery.get('power_plugged')),
        }


# vim:sw=4:ts=4:et:



import enum
import math
import time

from platypush.plugins import action
from platypush.plugins.gpio.sensor import GpioSensorPlugin


class Rotation(enum.IntEnum):
    ROTATE_0 = 0
    ROTATE_90 = 90
    ROTATE_180 = 180
    ROTATE_270 = 270


class SPISlot(enum.Enum):
    FRONT = 'front'
    BACK = 'back'


class GpioSensorMotionPwm3901Plugin(GpioSensorPlugin):
    """
    Plugin to interact with an `PWM3901 <https://github.com/pimoroni/pmw3901-python>`_
    optical flow and motion sensor

    Requires:

        * ``pwm3901`` (``pip install pwm3901``)
    """

    def __init__(self, rotation=Rotation.ROTATE_0.value, spi_slot=SPISlot.FRONT.value, spi_port=0, **kwargs):
        """
        :param rotation: Rotation angle for the captured optical flow. Possible options: 0, 90, 180, 270 (default: 0)
        :type rotation: int

        :param spi_slot: SPI slot where the sensor is connected if you're using a Breakout Garden interface.
            Possible options: 'front', 'back' (default: 'front')
        :type spi_slot: str

        :param spi_port: SPI port (default: 0)
        :type spi_slot: int
        """
        from pmw3901 import BG_CS_FRONT_BCM, BG_CS_BACK_BCM
        super().__init__(**kwargs)

        self.spi_port = spi_port
        self._sensor = None
        self._events_per_sec = {}
        self.x, self.y = (0, 0)

        try:
            if isinstance(rotation, int):
                rotation = [r for r in Rotation if r.value == rotation][0]
            self.rotation = rotation
        except IndexError:
            raise ValueError('{} is not a valid value for rotation - possible values: {}'.format(
                rotation, [r.value for r in Rotation]))

        try:
            if isinstance(spi_slot, str):
                spi_slot = [s for s in SPISlot if s.value == spi_slot][0]

            if spi_slot == SPISlot.FRONT:
                self.spi_slot = BG_CS_FRONT_BCM
            else:
                self.spi_slot = BG_CS_BACK_BCM
        except IndexError:
            raise ValueError('{} is not a valid value for spi_slot - possible values: {}'.format(
                spi_slot, [s.value for s in SPISlot]))

    def _get_sensor(self):
        from pmw3901 import PMW3901

        if not self._sensor:
            self._sensor = PMW3901(spi_port=self.spi_port,
                                   spi_cs=1,
                                   spi_cs_gpio=self.spi_slot)
            self._sensor.set_rotation(self.rotation)

        return self._sensor

    @action
    def get_measurement(self):
        """
        :returns: dict. Example:

        .. code-block:: python

           output = {
               "motion_x": 3,   # Detected motion vector X-coord
               "motion_y": 4,   # Detected motion vector Y-coord
               "motion_mod": 5  # Detected motion vector module
               "motion_events_per_sec": 7  # Number of motion events detected in the last second
           }

        """

        sensor = self._get_sensor()

        while True:
            try:
                x, y = sensor.get_motion()
                break
            except RuntimeError:
                time.sleep(0.01)

        secs = int(time.time())
        if (x, y) != (self.x, self.y):
            (self.x, self.y) = (x, y)

            if secs not in self._events_per_sec:
                self._events_per_sec = {secs: 1}
            else:
                self._events_per_sec[secs] += 1

        return {
            'motion_x': x,
            'motion_y': y,
            'motion_mod': math.sqrt(x * x + y * y),
            'motion_events_per_sec': self._events_per_sec.get(secs, 0),
        }


# vim:sw=4:ts=4:et:



from platypush.message.event import Event

class GoogleDeviceEvent(Event):
    """
    Base class for Google device events, see `managing traits and handlers <https://developers.google.com/assistant/sdk/guides/service/python/extend/add-trait-and-handler>`_.
    """

    def __init__(self, device_id, device_model_id=None, *args, **kwargs):
        super().__init__(*args, device_id=device_id, device_model_id=device_model_id, **kwargs)


class GoogleDeviceOnOffEvent(GoogleDeviceEvent):
    """
    Event triggered when a device receives an on/off command
    """

    def __init__(self, on, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.on = on


# vim:sw=4:ts=4:et:



"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

import time

from platypush.context import get_backend
from platypush.plugins import Plugin, action

class WiimotePlugin(Plugin):
    """
    WiiMote plugin.
    A wrapper around the :mod:`platypush.backend.wiimote` backend to
    programmatically control a Nintendo WiiMote.

    It requires the WiiMote backend to be enabled.
    """

    @classmethod
    def _get_wiimote(cls):
        return get_backend('wiimote').get_wiimote()


    @action
    def connect(self):
        """
        Connects to the WiiMote
        """
        self._get_wiimote()


    @action
    def close(self):
        """
        Closes the connection with the WiiMote
        """
        get_backend('wiimote').close()


    @action
    def rumble(self, secs):
        """
        Rumbles the controller for the specified number of seconds
        """
        wm = self._get_wiimote()
        wm.rumble = True
        time.sleep(secs)
        wm.rumble = False


    @action
    def state(self):
        """
        Return the state of the controller
        """
        return get_backend('wiimote').get_state()


    @action
    def set_leds(self, leds):
        """
        Set the LEDs state on the controller

        :param leds: Iterable with the new states to be applied to the LEDs. Example: [1, 0, 0, 0] or (False, True, False, False)
        :type leds: list
        """

        new_led = 0
        for i, led in enumerate(leds):
            if led:
                new_led |= (1 << i)

        self._get_wiimote().led = new_led


# vim:sw=4:ts=4:et:




from platypush.message.response import Response


class BluetoothResponse(Response):
    pass


class BluetoothScanResponse(BluetoothResponse):
    def __init__(self, devices, *args, **kwargs):
        if isinstance(devices, list):
            self.devices = [
                {
                    'addr': dev[0],
                    'name': dev[1] if len(dev) > 1 else None,
                    'class': hex(dev[2]) if len(dev) > 2 else None,
                }
                for dev in devices
            ]
        elif isinstance(devices, dict):
            self.devices = [
                {
                    'addr': addr,
                    'name': name or None,
                    'class': 'BLE',
                }
                for addr, name in devices.items()
            ]
        else:
            raise ValueError('devices must be either a list of tuples or a dict')

        super().__init__(output=self.devices, *args, **kwargs)


class BluetoothLookupNameResponse(BluetoothResponse):
    def __init__(self, addr: str, name: str, *args, **kwargs):
        self.addr = addr
        self.name = name
        super().__init__(output={
            'addr': self.addr,
            'name': self.name
        }, *args, **kwargs)


class BluetoothLookupServiceResponse(BluetoothResponse):
    """
    Example services response output::

        [
          {
            "service-classes": [
              "1801"
            ],
            "profiles": [],
            "name": "Service name #1",
            "description": null,
            "provider": null,
            "service-id": null,
            "protocol": "L2CAP",
            "port": 31,
            "host": "00:11:22:33:44:55"
          },
          {
            "service-classes": [
              "1800"
            ],
            "profiles": [],
            "name": "Service name #2",
            "description": null,
            "provider": null,
            "service-id": null,
            "protocol": "L2CAP",
            "port": 31,
            "host": "00:11:22:33:44:56"
          },
          {
            "service-classes": [
              "1112",
              "1203"
            ],
            "profiles": [
              [
                "1108",
                258
              ]
            ],
            "name": "Headset Gateway",
            "description": null,
            "provider": null,
            "service-id": null,
            "protocol": "RFCOMM",
            "port": 2,
            "host": "00:11:22:33:44:57"
          }
        ]
    """
    def __init__(self, services: list, *args, **kwargs):
        self.services = services
        super().__init__(output=self.services, *args, **kwargs)


class BluetoothDiscoverPrimaryResponse(BluetoothResponse):
    """
    Example services response output::

        [
          {
            "uuid": "00001800-0000-1000-8000-00805f9b34fb",
            "start": 1,
            "end": 7
          },
          {
            "uuid": "00001801-0000-1000-8000-00805f9b34fb",
            "start": 8,
            "end": 8
          },
          {
            "uuid": "0000fee7-0000-1000-8000-00805f9b34fb",
            "start": 9,
            "end": 16
          },
          {
            "uuid": "cba20d00-224d-11e6-9fb8-0002a5d5c51b",
            "start": 17,
            "end": 65535
          }
        ]

    """
    def __init__(self, services: list, *args, **kwargs):
        self.services = services
        super().__init__(output=self.services, *args, **kwargs)


class BluetoothDiscoverCharacteristicsResponse(BluetoothResponse):
    """
    Example services response output::

        [
          {
            "uuid": "00002a00-0000-1000-8000-00805f9b34fb",
            "handle": 2,
            "properties": 10,
            "value_handle": 3
          },
          {
            "uuid": "00002a01-0000-1000-8000-00805f9b34fb",
            "handle": 4,
            "properties": 2,
            "value_handle": 5
          },
          {
            "uuid": "00002a04-0000-1000-8000-00805f9b34fb",
            "handle": 6,
            "properties": 2,
            "value_handle": 7
          },
          {
            "uuid": "0000fec8-0000-1000-8000-00805f9b34fb",
            "handle": 10,
            "properties": 32,
            "value_handle": 11
          },
          {
            "uuid": "0000fec7-0000-1000-8000-00805f9b34fb",
            "handle": 13,
            "properties": 8,
            "value_handle": 14
          },
          {
            "uuid": "0000fec9-0000-1000-8000-00805f9b34fb",
            "handle": 15,
            "properties": 2,
            "value_handle": 16
          },
          {
            "uuid": "cba20003-224d-11e6-9fb8-0002a5d5c51b",
            "handle": 18,
            "properties": 16,
            "value_handle": 19
          },
          {
            "uuid": "cba20002-224d-11e6-9fb8-0002a5d5c51b",
            "handle": 21,
            "properties": 12,
            "value_handle": 22
          }
        ]

    """
    def __init__(self, characteristics: list, *args, **kwargs):
        self.characteristics = characteristics
        super().__init__(output=self.characteristics, *args, **kwargs)


# vim:sw=4:ts=4:et:



from typing import Dict, Any

from platypush.message.event import Event


class ZigbeeMqttEvent(Event):
    pass


class ZigbeeMqttOnlineEvent(ZigbeeMqttEvent):
    """
    Triggered when a zigbee2mqtt service goes online.
    """
    def __init__(self, host: str, port: int, *args, **kwargs):
        super().__init__(*args, host=host, port=port, **kwargs)


class ZigbeeMqttOfflineEvent(ZigbeeMqttEvent):
    """
    Triggered when a zigbee2mqtt service goes offline.
    """
    def __init__(self, host: str, port: int, *args, **kwargs):
        super().__init__(*args, host=host, port=port, **kwargs)


class ZigbeeMqttDevicePropertySetEvent(ZigbeeMqttEvent):
    """
    Triggered when a the properties of a Zigbee connected devices (state, brightness, alert etc.) change.
    """
    def __init__(self, host: str, port: int, device: str, properties: Dict[str, Any], *args, **kwargs):
        super().__init__(*args, host=host, port=port, device=device, properties=properties, **kwargs)


class ZigbeeMqttDevicePairingEvent(ZigbeeMqttEvent):
    """
    Triggered when a device is pairing to the network.
    """
    def __init__(self, host: str, port: int, device=None, *args, **kwargs):
        super().__init__(*args, host=host, port=port, device=device, **kwargs)


class ZigbeeMqttDeviceConnectedEvent(ZigbeeMqttEvent):
    """
    Triggered when a device connects to the network.
    """
    def __init__(self, host: str, port: int, device=None, *args, **kwargs):
        super().__init__(*args, host=host, port=port, device=device, **kwargs)


class ZigbeeMqttDeviceBannedEvent(ZigbeeMqttEvent):
    """
    Triggered when a device is banned from the network.
    """
    def __init__(self, host: str, port: int, device=None, *args, **kwargs):
        super().__init__(*args, host=host, port=port, device=device, **kwargs)


class ZigbeeMqttDeviceRemovedEvent(ZigbeeMqttEvent):
    """
    Triggered when a device is removed from the network.
    """
    def __init__(self, host: str, port: int, device=None, force=False, *args, **kwargs):
        super().__init__(*args, host=host, port=port, device=device, force=force, **kwargs)


class ZigbeeMqttDeviceRemovedFailedEvent(ZigbeeMqttEvent):
    """
    Triggered when the removal of a device from the network failed.
    """
    def __init__(self, host: str, port: int, device=None, *args, **kwargs):
        super().__init__(*args, host=host, port=port, device=device, **kwargs)


class ZigbeeMqttDeviceWhitelistedEvent(ZigbeeMqttEvent):
    """
    Triggered when a device is whitelisted on the network.
    """
    def __init__(self, host: str, port: int, device=None, *args, **kwargs):
        super().__init__(*args, host=host, port=port, device=device, **kwargs)


class ZigbeeMqttDeviceRenamedEvent(ZigbeeMqttEvent):
    """
    Triggered when a device is renamed on the network.
    """
    def __init__(self, host: str, port: int, device=None, *args, **kwargs):
        super().__init__(*args, host=host, port=port, device=device, **kwargs)


class ZigbeeMqttDeviceBindEvent(ZigbeeMqttEvent):
    """
    Triggered when a device bind occurs on the network.
    """
    def __init__(self, host: str, port: int, device=None, *args, **kwargs):
        super().__init__(*args, host=host, port=port, device=device, **kwargs)


class ZigbeeMqttDeviceUnbindEvent(ZigbeeMqttEvent):
    """
    Triggered when a device bind occurs on the network.
    """
    def __init__(self, host: str, port: int, device=None, *args, **kwargs):
        super().__init__(*args, host=host, port=port, device=device, **kwargs)


class ZigbeeMqttGroupAddedEvent(ZigbeeMqttEvent):
    """
    Triggered when a group is added.
    """
    def __init__(self, host: str, port: int, group=None, *args, **kwargs):
        super().__init__(*args, host=host, port=port, group=group, **kwargs)


class ZigbeeMqttGroupAddedFailedEvent(ZigbeeMqttEvent):
    """
    Triggered when a request to add a group fails.
    """
    def __init__(self, host: str, port: int, group=None, *args, **kwargs):
        super().__init__(*args, host=host, port=port, group=group, **kwargs)


class ZigbeeMqttGroupRemovedEvent(ZigbeeMqttEvent):
    """
    Triggered when a group is removed.
    """
    def __init__(self, host: str, port: int, group=None, *args, **kwargs):
        super().__init__(*args, host=host, port=port, group=group, **kwargs)


class ZigbeeMqttGroupRemovedFailedEvent(ZigbeeMqttEvent):
    """
    Triggered when a request to remove a group fails.
    """
    def __init__(self, host: str, port: int, group=None, *args, **kwargs):
        super().__init__(*args, host=host, port=port, group=group, **kwargs)


class ZigbeeMqttGroupRemoveAllEvent(ZigbeeMqttEvent):
    """
    Triggered when all the devices are removed from a group.
    """
    def __init__(self, host: str, port: int, group=None, *args, **kwargs):
        super().__init__(*args, host=host, port=port, group=group, **kwargs)


class ZigbeeMqttGroupRemoveAllFailedEvent(ZigbeeMqttEvent):
    """
    Triggered when a request to remove all the devices from a group fails.
    """
    def __init__(self, host: str, port: int, group=None, *args, **kwargs):
        super().__init__(*args, host=host, port=port, group=group, **kwargs)


class ZigbeeMqttErrorEvent(ZigbeeMqttEvent):
    """
    Triggered when an error happens on the zigbee2mqtt service.
    """
    def __init__(self, host: str, port: int, error=None, *args, **kwargs):
        super().__init__(*args, host=host, port=port, error=error, **kwargs)


# vim:sw=4:ts=4:et:



import datetime

from platypush.message.event import Event


class TrelloEvent(Event):
    pass


class CardEvent(TrelloEvent):
    def __init__(self,
                 card_id: str,
                 card_name: str,
                 list_id: str,
                 list_name: str,
                 board_id: str,
                 board_name: str,
                 closed: bool,
                 member_id: str,
                 member_username: str,
                 member_fullname: str,
                 date: datetime.datetime,
                 *args, **kwargs):
        super().__init__(*args,
                         card_id=card_id,
                         card_name=card_name,
                         list_id=list_id,
                         list_name=list_name,
                         board_id=board_id,
                         board_name=board_name,
                         closed=closed,
                         member_id=member_id,
                         member_username=member_username,
                         member_fullname=member_fullname,
                         date=date,
                         **kwargs)


class NewCardEvent(CardEvent):
    """
    Event triggered when a card is created.
    """


class MoveCardEvent(CardEvent):
    """
    Event triggered when a card is moved to another list.
    """

    def __init__(self, old_list_id: str, old_list_name: str, *args, **kwargs):
        super().__init__(*args, old_list_id=old_list_id, old_list_name=old_list_name, **kwargs)


class ArchivedCardEvent(CardEvent):
    """
    Event triggered when a card is archived.
    """

    def __init__(self, *args, **kwargs):
        kwargs['old_closed'] = False
        super().__init__(*args, **kwargs)



class UnarchivedCardEvent(CardEvent):
    """
    Event triggered when a card is un-archived.
    """

    def __init__(self, *args, **kwargs):
        kwargs['old_closed'] = True
        super().__init__(*args, **kwargs)


# vim:sw=4:ts=4:et:



import json

from platypush.message.event import Event


class NFCEvent(Event):
    """
    Generic class for NFC events
    """

    def __init__(self, reader=None, tag_id=None, *args, **kwargs):
        super().__init__(reader=reader, tag_id=tag_id, *args, **kwargs)


class NFCDeviceConnectedEvent(NFCEvent):
    """
    Event triggered when an NFC reader/writer devices is connected
    """

    def __init__(self, reader=None, *args, **kwargs):
        """
        :param reader: Name or address of the reader that fired the event
        :type reader: str
        """
        super().__init__(reader=reader, *args, **kwargs)


class NFCDeviceDisconnectedEvent(NFCEvent):
    """
    Event triggered when an NFC reader/writer devices is disconnected
    """

    def __init__(self, reader=None, *args, **kwargs):
        """
        :param reader: Name or address of the reader that fired the event
        :type reader: str
        """
        super().__init__(reader=reader, *args, **kwargs)


class NFCTagDetectedEvent(NFCEvent):
    """
    Event triggered when an NFC tag is connected
    """

    def __init__(self, reader=None, tag_id=None, records=None, *args, **kwargs):
        """
        :param reader: Name or address of the reader that fired the event
        :type reader: str

        :param tag_id: ID of the NFC tag
        :type tag_id: str

        :param records: Optional, list of records read from the tag. If the tag contains JSON-serializable data then it
            will be cast by the backend into the appropriate object
        :type records: str, bytes or JSON-serializable object
        """
        if not records:
            records = []

        super().__init__(reader=reader, tag_id=tag_id, records=records, *args, **kwargs)


class NFCTagRemovedEvent(NFCEvent):
    """
    Event triggered when a NFC card is removed/disconnected
    """

    def __init__(self, reader=None, tag_id=None, *args, **kwargs):
        """
        :param reader: Name or address of the reader that fired the event
        :type reader: str

        :param tag_id: ID of the NFC tag
        :type tag_id: str
        """
        super().__init__(reader=reader, tag_id=tag_id, *args, **kwargs)


# vim:sw=4:ts=4:et:



import os
import threading

from platypush.context import get_bus
from platypush.plugins.media import PlayerState, MediaPlugin
from platypush.message.event.media import MediaPlayEvent, MediaPlayRequestEvent, \
    MediaPauseEvent, MediaStopEvent, NewPlayingMediaEvent, MediaSeekEvent

from platypush.plugins import action


class MediaMpvPlugin(MediaPlugin):
    """
    Plugin to control MPV instances

    Requires:

        * **python-mpv** (``pip install python-mpv``)
        * **mpv** executable on your system
    """

    _default_mpv_args = {
        'ytdl': True,
        'start_event_thread': True,
    }

    def __init__(self, args=None, *argv, **kwargs):
        """
        Create the MPV wrapper.

        :param args: Default arguments that will be passed to the mpv executable
            as a key-value dict (names without the `--` prefix). See `man mpv`
            for available options.
        :type args: dict[str, str]
        """

        super().__init__(*argv, **kwargs)

        self.args = self._default_mpv_args
        if args:
            # noinspection PyTypeChecker
            self.args.update(args)

        self._player = None
        self._playback_rebounce_event = threading.Event()
        self._on_stop_callbacks = []

    def _init_mpv(self, args=None):
        import mpv

        mpv_args = self.args.copy()
        if args:
            mpv_args.update(args)

        for k, v in self._env.items():
            os.environ[k] = v

        self._player = mpv.MPV(**mpv_args)
        # noinspection PyProtectedMember
        self._player._event_callbacks += [self._event_callback()]

    @staticmethod
    def _post_event(evt_type, **evt):
        bus = get_bus()
        bus.post(evt_type(player='local', plugin='media.mpv', **evt))

    def _event_callback(self):
        def callback(event):
            from mpv import MpvEventID as Event
            from mpv import MpvEventEndFile as EndFile

            self.logger.info('Received mpv event: {}'.format(event))

            evt = event.get('event_id')
            if not evt:
                return

            if (evt == Event.FILE_LOADED or evt == Event.START_FILE) and self._get_current_resource():
                self._playback_rebounce_event.set()
                self._post_event(NewPlayingMediaEvent, resource=self._get_current_resource(),
                                 title=self._player.filename)
            elif evt == Event.PLAYBACK_RESTART:
                self._playback_rebounce_event.set()
            elif evt == Event.PAUSE:
                self._post_event(MediaPauseEvent, resource=self._get_current_resource(), title=self._player.filename)
            elif evt == Event.UNPAUSE:
                self._post_event(MediaPlayEvent, resource=self._get_current_resource(), title=self._player.filename)
            elif evt == Event.SHUTDOWN or evt == Event.IDLE or (
                    evt == Event.END_FILE and event.get('event', {}).get('reason') in
                    [EndFile.EOF_OR_INIT_FAILURE, EndFile.ABORTED, EndFile.QUIT]):
                playback_rebounced = self._playback_rebounce_event.wait(timeout=0.5)
                if playback_rebounced:
                    self._playback_rebounce_event.clear()
                    return

                self._player = None
                self._post_event(MediaStopEvent)

                for cbk in self._on_stop_callbacks:
                    cbk()
            elif evt == Event.SEEK:
                self._post_event(MediaSeekEvent, position=self._player.playback_time)

        return callback

    @action
    def execute(self, cmd, **args):
        """
        Execute a raw mpv command.
        """
        if not self._player:
            return None, 'No mpv instance is running'
        return self._player.command(cmd, *args)

    @action
    def play(self, resource, subtitles=None, **args):
        """
        Play a resource.

        :param resource: Resource to play - can be a local file or a remote URL
        :type resource: str

        :param subtitles: Path to optional subtitle file
        :type subtitles: str

        :param args: Extra runtime arguments that will be passed to the
            mpv executable as a key-value dict (keys without `--` prefix)
        :type args: dict[str,str]
        """

        get_bus().post(MediaPlayRequestEvent(resource=resource))
        self._init_mpv(args)

        if subtitles:
            args['sub_file'] = self.get_subtitles_file(subtitles)

        resource = self._get_resource(resource)

        if resource.startswith('file://'):
            resource = resource[7:]

        self._player.play(resource)
        if self.volume:
            self.set_volume(volume=self.volume)

        return self.status()

    @action
    def pause(self):
        """ Toggle the paused state """
        if not self._player:
            return None, 'No mpv instance is running'

        self._player.pause = not self._player.pause
        return self.status()

    @action
    def quit(self):
        """ Stop and quit the player """
        self._stop_torrent()
        if not self._player:
            return None, 'No mpv instance is running'

        self._player.quit()
        self._player.terminate()
        self._player = None
        return {'state': PlayerState.STOP.value}

    @action
    def stop(self):
        """ Stop and quit the player """
        return self.quit()

    @action
    def voldown(self, step=10.0):
        """ Volume down by (default: 10)% """
        if not self._player:
            return None, 'No mpv instance is running'
        return self.set_volume(self._player.volume - step)

    @action
    def volup(self, step=10.0):
        """ Volume up by (default: 10)% """
        if not self._player:
            return None, 'No mpv instance is running'
        return self.set_volume(self._player.volume + step)

    @action
    def set_volume(self, volume):
        """
        Set the volume

        :param volume: Volume value between 0 and 100
        :type volume: float
        """
        if not self._player:
            return None, 'No mpv instance is running'

        volume = max(0, min([self._player.volume_max, volume]))
        self._player.volume = volume
        return self.status()

    @action
    def seek(self, position):
        """
        Seek backward/forward by the specified number of seconds

        :param position: Number of seconds relative to the current cursor
        :type position: int
        """
        if not self._player:
            return None, 'No mpv instance is running'
        if not self._player.seekable:
            return None, 'The resource is not seekable'
        pos = min(self._player.time_pos + self._player.time_remaining,
                  max(0, position))
        self._player.time_pos = pos
        return self.status()

    @action
    def back(self, offset=60.0):
        """ Back by (default: 60) seconds """
        if not self._player:
            return None, 'No mpv instance is running'
        if not self._player.seekable:
            return None, 'The resource is not seekable'
        pos = max(0, self._player.time_pos - offset)
        return self.seek(pos)

    @action
    def forward(self, offset=60.0):
        """ Forward by (default: 60) seconds """
        if not self._player:
            return None, 'No mpv instance is running'
        if not self._player.seekable:
            return None, 'The resource is not seekable'
        pos = min(self._player.time_pos + self._player.time_remaining,
                  self._player.time_pos + offset)
        return self.seek(pos)

    @action
    def next(self):
        """ Play the next item in the queue """
        if not self._player:
            return None, 'No mpv instance is running'
        self._player.playlist_next()

    @action
    def prev(self):
        """ Play the previous item in the queue """
        if not self._player:
            return None, 'No mpv instance is running'
        self._player.playlist_prev()

    @action
    def toggle_subtitles(self, visible=None):
        """ Toggle the subtitles visibility """
        return self.toggle_property('sub_visibility')

    @action
    def add_subtitles(self, filename):
        """ Add a subtitles file """
        return self._player.sub_add(filename)

    @action
    def remove_subtitles(self, sub_id):
        """ Remove a subtitles track by id """
        return self._player.sub_remove(sub_id)

    @action
    def toggle_fullscreen(self):
        """ Toggle the fullscreen mode """
        return self.toggle_property('fullscreen')

    # noinspection PyShadowingBuiltins
    @action
    def toggle_property(self, property):
        """
        Toggle or sets the value of an mpv property (e.g. fullscreen,
        sub_visibility etc.). See ``man mpv`` for a full list of properties

        :param property: Property to toggle
        """
        if not self._player:
            return None, 'No mpv instance is running'

        if not hasattr(self._player, property):
            self.logger.warning('No such mpv property: {}'.format(property))

        value = not getattr(self._player, property)
        setattr(self._player, property, value)
        return {property: value}

    # noinspection PyShadowingBuiltins
    @action
    def get_property(self, property):
        """
        Get a player property (e.g. pause, fullscreen etc.). See
        ``man mpv`` for a full list of the available properties
        """
        if not self._player:
            return None, 'No mpv instance is running'
        return getattr(self._player, property)

    @action
    def set_property(self, **props):
        """
        Set the value of an mpv property (e.g. fullscreen, sub_visibility
        etc.). See ``man mpv`` for a full list of properties

        :param props: Key-value args for the properties to set
        :type props: dict
        """
        if not self._player:
            return None, 'No mpv instance is running'

        for k, v in props.items():
            setattr(self._player, k, v)
        return props

    @action
    def set_subtitles(self, filename, *args, **kwargs):
        """ Sets media subtitles from filename """
        # noinspection PyTypeChecker
        return self.set_property(subfile=filename, sub_visibility=True)

    @action
    def remove_subtitles(self):
        """ Removes (hides) the subtitles """
        if not self._player:
            return None, 'No mpv instance is running'
        self._player.sub_visibility = False

    @action
    def is_playing(self):
        """
        :returns: True if it's playing, False otherwise
        """
        if not self._player:
            return False
        return not self._player.pause

    @action
    def load(self, resource, **args):
        """
        Load/queue a resource/video to the player
        """
        if not self._player:
            return self.play(resource, **args)
        return self._player.loadfile(resource, mode='append-play')

    @action
    def mute(self):
        """ Toggle mute state """
        if not self._player:
            return None, 'No mpv instance is running'
        mute = not self._player.mute
        self._player.mute = mute
        return {'muted': mute}

    @action
    def set_position(self, position):
        """
        Seek backward/forward to the specified absolute position (same as ``seek``)
        """
        return self.seek(position)

    @action
    def status(self):
        """
        Get the current player state.

        :returns: A dictionary containing the current state.

        Example::

            output = {
                "filename": "filename or stream URL",
                "state": "play"  # or "stop" or "pause"
            }
        """
        if not self._player or not hasattr(self._player, 'pause'):
            return {'state': PlayerState.STOP.value}

        return {
            'audio_channels': getattr(self._player, 'audio_channels'),
            'audio_codec': getattr(self._player, 'audio_codec_name'),
            'delay': getattr(self._player, 'delay'),
            'duration': getattr(self._player, 'playback_time', 0) + getattr(self._player, 'playtime_remaining', 0)
            if getattr(self._player, 'playtime_remaining') else None,
            'filename': getattr(self._player, 'filename'),
            'file_size': getattr(self._player, 'file_size'),
            'fullscreen': getattr(self._player, 'fs'),
            'mute': getattr(self._player, 'mute'),
            'name': getattr(self._player, 'name'),
            'pause': getattr(self._player, 'pause'),
            'percent_pos': getattr(self._player, 'percent_pos'),
            'position': getattr(self._player, 'playback_time'),
            'seekable': getattr(self._player, 'seekable'),
            'state': (PlayerState.PAUSE.value if self._player.pause else PlayerState.PLAY.value),
            'title': getattr(self._player, 'media_title') or getattr(self._player, 'filename'),
            'url': self._get_current_resource(),
            'video_codec': getattr(self._player, 'video_codec'),
            'video_format': getattr(self._player, 'video_format'),
            'volume': getattr(self._player, 'volume'),
            'volume_max': getattr(self._player, 'volume_max'),
            'width': getattr(self._player, 'width'),
        }

    def on_stop(self, callback):
        self._on_stop_callbacks.append(callback)

    def _get_current_resource(self):
        if not self._player or not self._player.stream_path:
            return

        return ('file://' if os.path.isfile(self._player.stream_path)
                else '') + self._player.stream_path

# vim:sw=4:ts=4:et:



from typing import Dict, Optional, List

from platypush.backend.sensor import SensorBackend
from platypush.message.event.linode import LinodeInstanceStatusChanged


class LinodeBackend(SensorBackend):
    """
    This backend monitors the state of one or more Linode instances.

    Triggers:

        * :class:`platypush.message.event.linode.LinodeInstanceStatusChanged` when the status of an instance changes.

    Requires:

        * The :class:`platypush.plugins.linode.LinodePlugin` plugin configured.

    """

    def __init__(self, instances: Optional[List[str]] = None, poll_seconds: float = 30.0, **kwargs):
        """
        :param instances: List of instances to monitor, by label (default: monitor all the instances).
        """
        super().__init__(plugin='linode', poll_seconds=poll_seconds, **kwargs)
        self.instances = set(instances or [])

    def process_data(self, data: Dict[str, dict], *args, **kwargs):
        instances = data['instances']
        old_instances = (self.data or {}).get('instances', {})

        if self.instances:
            instances = {label: instances[label] for label in self.instances if label in instances}

        if not instances:
            return

        for label, instance in instances.items():
            old_instance = old_instances.get(label, {})
            if 'status' in old_instance and old_instance['status'] != instance['status']:
                self.bus.post(LinodeInstanceStatusChanged(instance=label,
                                                          status=instance['status'],
                                                          old_status=old_instance['status']))


# vim:sw=4:ts=4:et:



import base64
import io
import os
import threading
import time
from typing import Optional, List

import numpy as np
from PIL import Image

from platypush import Config
from platypush.context import get_bus
from platypush.message.event.qrcode import QrcodeScannedEvent
from platypush.message.response.qrcode import QrcodeGeneratedResponse, QrcodeDecodedResponse, ResultModel
from platypush.plugins import Plugin, action
from platypush.plugins.camera import CameraPlugin
from platypush.utils import get_plugin_class_by_name


class QrcodePlugin(Plugin):
    """
    Plugin to generate and scan QR and bar codes.

    Requires:

        * **numpy** (``pip install numpy``).
        * **qrcode** (``pip install 'qrcode[pil]'``) for QR generation.
        * **pyzbar** (``pip install pyzbar``) for decoding code from images.
        * **Pillow** (``pip install Pillow``) for image management.

    """

    def __init__(self, camera_plugin: Optional[str] = None, **kwargs):
        """
        :param camera_plugin: Name of the plugin that will be used as a camera to capture images (e.g.
            ``camera`` or ``camera.pi``).
        """
        super().__init__(**kwargs)
        self.camera_plugin = camera_plugin
        self._capturing = threading.Event()

    def _get_camera(self, camera_plugin: Optional[str] = None, **config) -> CameraPlugin:
        camera_plugin = camera_plugin or self.camera_plugin
        if not config:
            config = Config.get(camera_plugin) or {}
        config['stream_raw_frames'] = True

        cls = get_plugin_class_by_name(camera_plugin)
        assert cls and issubclass(cls, CameraPlugin), '{} is not a valid camera plugin'.format(camera_plugin)
        return cls(**config)

    # noinspection PyShadowingBuiltins
    @action
    def generate(self, content: str, output_file: Optional[str] = None, show: bool = False,
                 format: str = 'png', camera_plugin: Optional[str] = None) -> QrcodeGeneratedResponse:
        """
        Generate a QR code.
        If you configured the :class`:platypush.backend.http.HttpBackend` then you can also generate
        codes directly from the browser through ``http://<host>:<port>/qrcode?content=...``.

        :param content: Text, URL or content of the QR code.
        :param output_file: If set then the QR code will be exported in the specified image file.
            Otherwise, a base64-encoded representation of its binary content will be returned in
            the response as ``data``.
        :param show: If True, and if the device where the application runs has an active display,
            then the generated QR code will be shown on display.
        :param format: Output image format (default: ``png``).
        :param camera_plugin: If set then this plugin (e.g. ``camera`` or ``camera.pi``) will be used to capture
            live images from the camera and search for bar codes or QR-codes.
        :return: :class:`platypush.message.response.qrcode.QrcodeGeneratedResponse`.
        """
        import qrcode
        qr = qrcode.make(content)
        img = qr.get_image()
        ret = {
            'content': content,
            'format': format,
        }

        if show:
            img.show()
        if output_file:
            output_file = os.path.abspath(os.path.expanduser(output_file))
            img.save(output_file, format=format)
            ret['image_file'] = output_file
        else:
            f = io.BytesIO()
            img.save(f, format=format)
            ret['data'] = base64.encodebytes(f.getvalue()).decode()

        return QrcodeGeneratedResponse(**ret)

    @action
    def decode(self, image_file: str) -> QrcodeDecodedResponse:
        """
        Decode a QR code from an image file.

        :param image_file: Path of the image file.
        """
        from pyzbar import pyzbar
        from PIL import Image

        image_file = os.path.abspath(os.path.expanduser(image_file))
        img = Image.open(image_file)
        results = pyzbar.decode(img)
        return QrcodeDecodedResponse(results)

    def _convert_frame(self, frame) -> Image:
        assert isinstance(frame, np.ndarray), \
                'Image conversion only works with numpy arrays for now (got {})'.format(type(frame))
        mode = 'RGB'
        if len(frame.shape) > 2 and frame.shape[2] == 4:
            mode = 'RGBA'

        return Image.frombuffer(mode, (frame.shape[1], frame.shape[0]), frame, 'raw', mode, 0, 1)

    @action
    def start_scanning(self, camera_plugin: Optional[str] = None, duration: Optional[float] = None,
                       n_codes: Optional[int] = None) -> Optional[List[ResultModel]]:
        """
        Decode QR-codes and bar codes using a camera.

        Triggers:

            - :class:`platypush.message.event.qrcode.QrcodeScannedEvent` when a code is successfully scanned.

        :param camera_plugin: Camera plugin (overrides default ``camera_plugin``).
        :param duration: How long the capturing phase should run (default: until ``stop_scanning`` or app termination).
        :param n_codes: Stop after decoding this number of codes (default: None).
        :return: When ``duration`` or ``n_codes`` are specified or ``stop_scanning`` is called, it will return a list of
            :class:`platypush.message.response.qrcode.ResultModel` instances with the scanned results,
        """
        from pyzbar import pyzbar
        assert not self._capturing.is_set(), 'A capturing process is already running'

        camera = self._get_camera(camera_plugin)
        codes = []
        last_results = {}
        last_results_timeout = 10.0
        last_results_time = 0
        self._capturing.set()

        try:
            with camera:
                start_time = time.time()

                while self._capturing.is_set() \
                        and (not duration or time.time() < start_time + duration) \
                        and (not n_codes or len(codes) < n_codes):
                    output = camera.get_stream()
                    with output.ready:
                        output.ready.wait()
                        img = self._convert_frame(output.raw_frame)
                        results = pyzbar.decode(img)
                        if results:
                            results = [
                                result for result in QrcodeDecodedResponse(results).output['results']
                                if result['data'] not in last_results
                                   or time.time() >= last_results_time + last_results_timeout
                            ]

                        if results:
                            codes.extend(results)
                            get_bus().post(QrcodeScannedEvent(results=results))
                            last_results = {result['data']: result for result in results}
                            last_results_time = time.time()
        finally:
            self._capturing.clear()

        return codes

    @action
    def stop_scanning(self):
        self._capturing.clear()


# vim:sw=4:ts=4:et:



"""
Platydock

Platydock is a helper that allows you to easily manage (create, destroy, start,
stop and list) Platypush instances as Docker images.

.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
.. license: MIT
"""

import argparse
import enum
import os
import re
import subprocess
import sys
import textwrap
import traceback as tb

from platypush.config import Config
from platypush.context import register_backends, get_plugin, get_backend

workdir = os.path.join(os.path.expanduser('~'), '.local', 'share',
                       'platypush', 'platydock')


class Action(enum.Enum):
    build = 'build'
    start = 'start'
    stop = 'stop'
    rm = 'rm'
    ls = 'ls'

    def __str__(self):
        return self.value


def _parse_deps(cls):
    deps = []

    for line in cls.__doc__.split('\n'):
        m = re.search('\(``pip install (.+)``\)', line)
        if m:
            deps.append(m.group(1))

    return deps


def generate_dockerfile(deps, ports, cfgfile, devdir, python_version):
    device_id = Config.get('device_id')
    if not device_id:
        raise RuntimeError(('You need to specify a device_id in {} - Docker ' +
                            'containers cannot rely on hostname').format(cfgfile))

    os.makedirs(devdir, exist_ok=True)
    content = textwrap.dedent(
        '''
        FROM python:{python_version}-slim-buster

        RUN mkdir -p /app
        RUN mkdir -p /etc/platypush
        RUN mkdir -p /usr/local/share/platypush\n
        '''.format(python_version=python_version)).lstrip()

    srcdir = os.path.dirname(cfgfile)
    cfgfile_copy = os.path.join(devdir, 'config.yaml')
    subprocess.call(['cp', cfgfile, cfgfile_copy])
    content += 'COPY config.yaml /etc/platypush/\n'

    # noinspection PyProtectedMember
    for include in Config._included_files:
        incdir = os.path.relpath(os.path.dirname(include), srcdir)
        destdir = os.path.join(devdir, incdir)

        try:
            os.makedirs(destdir)
        except FileExistsError:
            pass

        subprocess.call(['cp', include, destdir])
        content += 'RUN mkdir -p /etc/platypush/' + incdir + '\n'
        content += 'COPY ' + os.path.relpath(include, srcdir) + \
                   ' /etc/platypush/' + incdir + '\n'

    content += textwrap.dedent(
        '''
        RUN dpkg --configure -a \\
            && apt-get -f install \\
            && apt-get --fix-missing install \\
            && apt-get clean \\
            && apt-get update \\
            && apt-get -y upgrade \\
            && apt-get -y dist-upgrade \\
            && apt-get install --no-install-recommends -y apt-utils\\
            && apt-get install --no-install-recommends -y build-essential \\
            && apt-get install --no-install-recommends -y git \\
            && apt-get install --no-install-recommends -y libffi-dev \\
            && apt-get install --no-install-recommends -y libjpeg-dev \\
            && apt-get install --no-install-recommends -y zlib1g-dev \\
        ''')

    for i, dep in enumerate(deps):
        content += '\t&& pip install --no-cache-dir {}'.format(dep)
        if i < len(deps)-1:
            content += ' \\'.format(dep)
        content += '\n'

    content += textwrap.dedent(
        '''

        RUN git clone --recursive https://github.com/BlackLight/platypush.git /app \\
            && cd /app \\
            && pip install -r requirements.txt \\
            && python setup.py web_build

        RUN apt-get remove -y git \\
            && apt-get remove -y build-essential \\
            && apt-get remove -y libffi-dev \\
            && apt-get remove -y libjpeg-dev \\
            && apt-get remove -y zlib1g-dev \\
            && apt-get remove -y apt-utils \\
            && apt-get clean \\
            && apt-get autoremove -y \\
            && rm -rf /var/lib/apt/lists/*

        ''')

    for port in ports:
        content += 'EXPOSE {}\n'.format(port)

    content += textwrap.dedent(
        '''

        ENV PYTHONPATH /app:$PYTHONPATH
        CMD ["python", "-m", "platypush"]
        ''')

    dockerfile = os.path.join(devdir, 'Dockerfile')
    print('Generating Dockerfile {}'.format(dockerfile))

    with open(dockerfile, 'w') as f:
        f.write(content)


def build(args):
    global workdir

    ports = set()
    deps = set()

    parser = argparse.ArgumentParser(prog='platydock build',
                                     description='Build a Platypush image ' +
                                                 'from a config.yaml')

    parser.add_argument('-c', '--config', type=str, required=True,
                        help='Path to the platypush configuration file')
    parser.add_argument('-p', '--python-version', type=str, default='3.8',
                        help='Python version to be used')

    opts, args = parser.parse_known_args(args)

    cfgfile = os.path.abspath(os.path.expanduser(opts.config))
    python_version = opts.python_version
    Config.init(cfgfile)
    register_backends()
    backend_config = Config.get_backends()

    if backend_config.get('http'):
        http_backend = get_backend('http')
        ports.add(http_backend.port)
        if http_backend.websocket_port:
            ports.add(http_backend.websocket_port)

    if backend_config.get('tcp'):
        ports.add(get_backend('tcp').port)

    if backend_config.get('websocket'):
        ports.add(get_backend('websocket').port)

    for name in Config.get_backends().keys():
        deps.update(_parse_deps(get_backend(name)))

    for name in Config.get_plugins().keys():
        try:
            deps.update(_parse_deps(get_plugin(name)))
        except Exception as ex:
            print('Dependencies parsing error for {}: {}'.format(name, str(ex)))

    devdir = os.path.join(workdir, Config.get('device_id'))
    generate_dockerfile(deps=deps, ports=ports, cfgfile=cfgfile, devdir=devdir, python_version=python_version)

    subprocess.call(['docker', 'build', '-t', 'platypush-{}'.format(
        Config.get('device_id')), devdir])


def start(args):
    global workdir

    parser = argparse.ArgumentParser(prog='platydock start',
                                     description='Start a Platypush container',
                                     epilog=textwrap.dedent('''
                                       You can append additional options that
                                       will be passed to the docker container.
                                       Example:

                                            --add-host='myhost:192.168.1.1'
                                       '''))

    parser.add_argument('image', type=str, help='Platypush image to start')
    parser.add_argument('-p', '--publish', action='append', nargs='*', default=[],
                        help=textwrap.dedent('''
                                             Container's ports to expose to the host.
                                             Note that the default exposed ports from
                                             the container service will be exposed unless
                                             these mappings override them (e.g. port 8008
                                             on the container will be mapped to 8008 on
                                             the host).

                                             Example:

                                             -p 18008:8008 -p 18009:8009
                                             '''))

    parser.add_argument('-a', '--attach', action='store_true', default=False,
                        help=textwrap.dedent('''
                                             If set, then attach to the container after starting it up (default: false).
                                             '''))


    opts, args = parser.parse_known_args(args)
    ports = {}
    dockerfile = os.path.join(workdir, opts.image, 'Dockerfile')

    with open(dockerfile) as f:
        for line in f:
            m = re.match('expose (\d+)', line.strip().lower())
            if m:
                ports[m.group(1)] = m.group(1)

    for mapping in opts.publish:
        host_port, container_port = mapping[0].split(':')
        ports[container_port] = host_port

    print('Preparing Redis support container')
    subprocess.call(['docker', 'pull', 'redis'])
    subprocess.call(['docker', 'run', '--rm', '--name', 'redis-' + opts.image,
                     '-d', 'redis'])

    docker_cmd = ['docker', 'run', '--rm', '--name', opts.image, '-it',
                  '--link', 'redis-' + opts.image + ':redis']

    for container_port, host_port in ports.items():
        docker_cmd += ['-p', host_port + ':' + container_port]

    docker_cmd += args
    docker_cmd += ['-d', 'platypush-' + opts.image]

    print('Starting Platypush container {}'.format(opts.image))
    subprocess.call(docker_cmd)

    if opts.attach:
        subprocess.call(['docker', 'attach', opts.image])


def stop(args):
    parser = argparse.ArgumentParser(prog='platydock stop',
                                     description='Stop a Platypush container')

    parser.add_argument('container', type=str, help='Platypush container to stop')
    opts, args = parser.parse_known_args(args)

    print('Stopping Platypush container {}'.format(opts.container))
    subprocess.call(['docker', 'kill', '{}'.format(opts.container)])

    print('Stopping Redis support container')
    subprocess.call(['docker', 'stop', 'redis-{}'.format(opts.container)])


def rm(args):
    global workdir

    parser = argparse.ArgumentParser(prog='platydock rm',
                                     description='Remove a Platypush image. ' +
                                                 'NOTE: make sure that no container is ' +
                                                 'running nor linked to the image before ' +
                                                 'removing it')

    parser.add_argument('image', type=str, help='Platypush image to remove')
    opts, args = parser.parse_known_args(args)

    subprocess.call(['docker', 'rmi', 'platypush-{}'.format(opts.image)])
    subprocess.call(['rm', '-r', os.path.join(workdir, opts.image)])


def ls(args):
    parser = argparse.ArgumentParser(prog='platydock ls',
                                     description='List available Platypush containers')
    parser.add_argument('filter', type=str, nargs='?',
                        help='Image name filter')

    opts, args = parser.parse_known_args(args)

    p = subprocess.Popen(['docker', 'images'], stdout=subprocess.PIPE)
    output = p.communicate()[0].decode().split('\n')
    header = output.pop(0)
    images = []

    for line in output:
        if re.match('^platypush-(.+?)\s.*', line):
            if not opts.filter or (opts.filter and opts.filter in line):
                images.append(line)

    if images:
        print(header)

        for image in images:
            print(image)


def main():
    parser = argparse.ArgumentParser(prog='platydock', add_help=False,
                                     description='Manage Platypush docker containers',
                                     epilog='Use platydock <action> --help to ' +
                                            'get additional help')

    # noinspection PyTypeChecker
    parser.add_argument('action', nargs='?', type=Action, choices=list(Action),
                        help='Action to execute')
    parser.add_argument('-h', '--help', action='store_true', help='Show usage')
    opts, args = parser.parse_known_args(sys.argv[1:])

    if (opts.help and not opts.action) or (not opts.help and not opts.action):
        parser.print_help()
        return 1

    globals()[str(opts.action)](sys.argv[2:])


if __name__ == '__main__':
    ERR_PREFIX = '\n\033[6;31;47mERROR\033[0;91m '
    ERR_SUFFIX = '\033[0m'

    try:
        main()
    except Exception as e:
        tb.print_exc(file=sys.stdout)
        print(ERR_PREFIX + str(e) + ERR_SUFFIX, file=sys.stderr)

# vim:sw=4:ts=4:et:



from platypush.plugins import action
from platypush.plugins.gpio.sensor import GpioSensorPlugin


class GpioSensorLtr559Plugin(GpioSensorPlugin):
    """
    Plugin to interact with an `LTR559 <https://shop.pimoroni.com/products/ltr-559-light-proximity-sensor-breakout>`_
    light and proximity sensor

    Requires:

        * ``ltr559`` (``pip install ltr559``)
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    # noinspection PyUnresolvedReferences
    @action
    def get_measurement(self):
        """
        :returns: dict. Example:

        .. code-block:: python
:
        output = {
            "light": 109.3543,     # Lux
            "proximity": 103       # The higher the value, the nearest the object, within a ~5cm range
        }

        """

        import ltr559
        ltr559.set_proximity_active()

        return {
            'light': ltr559.get_lux(),
            'proximity': ltr559.get_proximity(),
        }


# vim:sw=4:ts=4:et:



import logging

from functools import wraps

from platypush.event import EventGenerator
from platypush.message.response import Response
from platypush.utils import get_decorators


def action(f):
    @wraps(f)
    def _execute_action(*args, **kwargs):
        response = Response()
        result = f(*args, **kwargs)

        if result and isinstance(result, Response):
            result.errors = result.errors \
                if isinstance(result.errors, list) else [result.errors]
            response = result
        elif isinstance(result, tuple) and len(result) == 2:
            response.errors = result[1] \
                if isinstance(result[1], list) else [result[1]]

            if len(response.errors) == 1 and response.errors[0] is None:
                response.errors = []
            response.output = result[0]
        else:
            response = Response(output=result, errors=[])

        return response

    # Propagate the docstring
    _execute_action.__doc__ = f.__doc__
    return _execute_action


class Plugin(EventGenerator):
    """ Base plugin class """

    def __init__(self, **kwargs):
        super().__init__()

        self.logger = logging.getLogger(self.__class__.__name__)
        if 'logging' in kwargs:
            self.logger.setLevel(getattr(logging, kwargs['logging'].upper()))

        self.registered_actions = set(
            get_decorators(self.__class__, climb_class_hierarchy=True).get('action', [])
        )

    def run(self, method, *args, **kwargs):
        assert method in self.registered_actions, '{} is not a registered action on {}'.\
            format(method, self.__class__.__name__)
        return getattr(self, method)(*args, **kwargs)


# vim:sw=4:ts=4:et:



"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

import os
import socket
import threading
import time

from typing import Optional

from platypush.plugins import action
from platypush.plugins.camera import CameraPlugin, StreamingOutput


class CameraPiPlugin(CameraPlugin):
    """
    Plugin to control a Pi camera.

    Requires:

        * **picamera** (``pip install picamera``)
        * **numpy** (``pip install numpy``)

    """

    _default_resolution = (800, 600)
    _default_listen_port = 5000

    def __init__(self, resolution=(_default_resolution[0], _default_resolution[1]), framerate=24,
                 hflip=False, vflip=False, sharpness=0, contrast=0, brightness=50, video_stabilization=False, iso=0,
                 exposure_compensation=0, exposure_mode='auto', meter_mode='average', awb_mode='auto',
                 image_effect='none', color_effects=None, rotation=0, crop=(0.0, 0.0, 1.0, 1.0),
                 listen_port: int = _default_listen_port, **kwargs):
        """
        See https://www.raspberrypi.org/documentation/usage/camera/python/README.md
        for a detailed reference about the Pi camera options.

        :param listen_port: Default port that will be used for streaming the feed (default: 5000)
        """
        super().__init__(**kwargs)

        self.camera_args = {
            'resolution': tuple(resolution),
            'framerate': framerate,
            'hflip': hflip,
            'vflip': vflip,
            'sharpness': sharpness,
            'contrast': contrast,
            'brightness': brightness,
            'video_stabilization': video_stabilization,
            'ISO': iso,
            'exposure_compensation': exposure_compensation,
            'exposure_mode': exposure_mode,
            'meter_mode': meter_mode,
            'awb_mode': awb_mode,
            'image_effect': image_effect,
            'color_effects': color_effects,
            'rotation': rotation,
            'crop': tuple(crop),
        }

        self._camera = None
        self.listen_port = listen_port

        self._time_lapse_thread = None
        self._recording_thread = None
        self._streaming_thread = None
        self._capturing_thread = None
        self._time_lapse_stop_condition = threading.Condition()
        self._recording_stop_condition = threading.Condition()
        self._can_stream = False
        self._can_capture = False

    # noinspection PyUnresolvedReferences,PyPackageRequirements
    def _get_camera(self, **opts):
        if self._camera and not self._camera.closed:
            return self._camera

        import picamera
        self._camera = picamera.PiCamera()

        for (attr, value) in self.camera_args.items():
            setattr(self._camera, attr, value)
        for (attr, value) in opts.items():
            setattr(self._camera, attr, value)

        return self._camera

    @action
    def close(self):
        """
        Close an active connection to the camera.
        """
        import picamera

        if self._output and self._camera:
            try:
                self._camera.stop_recording()
            except picamera.PiCameraNotRecording:
                pass

        if self._camera and not self._camera.closed:
            try:
                self._camera.close()
            except picamera.PiCameraClosed:
                pass

        self._camera = None

    @action
    def start_preview(self, **opts):
        """
        Start camera preview.

        :param opts: Extra options to pass to the camera (see
            https://www.raspberrypi.org/documentation/usage/camera/python/README.md)
        """
        camera = self._get_camera(**opts)
        camera.start_preview()

    @action
    def stop_preview(self):
        """
        Stop camera preview.
        """
        camera = self._get_camera()
        try:
            camera.stop_preview()
        except Exception as e:
            self.logger.warning(str(e))

    @action
    def take_picture(self, image_file, preview=False, warmup_time=2, resize=None, close=True, **opts):
        """
        Take a picture.

        :param image_file: Path where the output image will be stored.
        :type image_file: str

        :param preview: Show a preview before taking the picture (default: False)
        :type preview: bool

        :param warmup_time: Time before taking the picture (default: 2 seconds)
        :type warmup_time: float

        :param resize: Set if you want to resize the picture to a new format
        :type resize: list or tuple (with two elements)

        :param opts: Extra options to pass to the camera (see
            https://www.raspberrypi.org/documentation/usage/camera/python/README.md)

        :param close: If True (default) close the connection to the camera after capturing,
            otherwise keep the connection open (e.g. if you want to take a sequence of pictures).
            If you set close=False you should remember to call ``close`` when you don't need
            the connection anymore.

        :return: dict::

            {"image_file": path_to_the_image}

        """

        camera = None

        try:
            camera = self._get_camera(**opts)
            image_file = os.path.abspath(os.path.expanduser(image_file))

            if preview:
                camera.start_preview()

            if warmup_time:
                time.sleep(warmup_time)

            capture_opts = {}
            if resize:
                capture_opts['resize'] = tuple(resize)

            camera.capture(image_file, **capture_opts)

            if preview:
                camera.stop_preview()

            return {'image_file': image_file}
        finally:
            if camera and close:
                self.close()

    def _raw_capture(self):
        import numpy as np
        resolution = self.camera_args['resolution']
        camera = self._get_camera()

        while self._can_capture:
            shape = (resolution[1] + (resolution[1]%16),
                     resolution[0] + (resolution[0]%32),
                     3)

            frame = np.empty(shape, dtype=np.uint8)
            camera.capture(frame, 'bgr')
            frame.reshape((shape[0], shape[1], 3))
            self._output.write(frame)

    def __enter__(self):
        camera = self._get_camera()
        self._output = StreamingOutput(raw=self.stream_raw_frames)
        self._can_capture = True

        if self.stream_raw_frames:
            self._capturing_thread = threading.Thread(target=self._raw_capture)
            self._capturing_thread.start()
        else:
            camera.start_recording(self._output, format='mjpeg')

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._can_capture = False
        if self._capturing_thread:
            self._capturing_thread.join()
            self._capturing_thread = None

        self.close()

    @action
    def capture_sequence(self, n_images, directory, name_format='image_%04d.jpg', preview=False, warmup_time=2,
                         resize=None, **opts):
        """
        Capture a sequence of images

        :param n_images: Number of images to capture
        :type n_images: int

        :param directory: Path where the images will be stored
        :type directory: str

        :param name_format: Format for the name of the stored images. Use %d or any other format string for representing
            the image index (default: image_%04d.jpg)
        :type name_format: str

        :param preview: Show a preview before taking the picture (default: False)
        :type preview: bool

        :param warmup_time: Time before taking the picture (default: 2 seconds)
        :type warmup_time: float

        :param resize: Set if you want to resize the picture to a new format
        :type resize: list or tuple (with two elements)

        :param opts: Extra options to pass to the camera (see
            https://www.raspberrypi.org/documentation/usage/camera/python/README.md)

        :return: dict::

            {"image_files": [list of captured images]}

        """

        try:
            camera = self._get_camera(**opts)
            directory = os.path.abspath(os.path.expanduser(directory))

            if preview:
                camera.start_preview()

            if warmup_time:
                time.sleep(warmup_time)
                camera.exposure_mode = 'off'

            camera.shutter_speed = camera.exposure_speed
            g = camera.awb_gains
            camera.awb_mode = 'off'
            camera.awb_gains = g
            capture_opts = {}

            if resize:
                capture_opts['resize'] = tuple(resize)

            images = [os.path.join(directory, name_format % (i+1)) for i in range(0, n_images)]
            camera.capture_sequence(images, **capture_opts)

            if preview:
                camera.stop_preview()

            return {'image_files': images}
        finally:
            self.close()

    @action
    def start_time_lapse(self, directory, n_images=None, interval=0, warmup_time=2,
                         resize=None, **opts):
        """
        Start a time lapse capture

        :param directory: Path where the images will be stored
        :type directory: str

        :param n_images: Number of images to capture (default: None, capture until stop_time_lapse)
        :type n_images: int

        :param interval: Interval in seconds between two pictures (default: 0)
        :type interval: float

        :param warmup_time: Time before taking the picture (default: 2 seconds)
        :type warmup_time: float

        :param resize: Set if you want to resize the picture to a new format
        :type resize: list or tuple (with two elements)

        :param opts: Extra options to pass to the camera (see
            https://www.raspberrypi.org/documentation/usage/camera/python/README.md)
        """

        if self._time_lapse_thread:
            return None, 'A time lapse thread is already running'

        camera = self._get_camera(**opts)
        directory = os.path.abspath(os.path.expanduser(directory))

        if warmup_time:
            time.sleep(warmup_time)

        capture_opts = {}
        if resize:
            capture_opts['resize'] = tuple(resize)

        def capture_thread():
            try:
                self.logger.info('Starting time lapse recording to directory {}'.format(directory))
                i = 0

                for filename in camera.capture_continuous(os.path.join(directory, 'image_{counter:04d}.jpg')):
                    i += 1
                    self.logger.info('Captured {}'.format(filename))

                    if n_images and i >= n_images:
                        break

                    self._time_lapse_stop_condition.acquire()
                    should_stop = self._time_lapse_stop_condition.wait(timeout=interval)
                    self._time_lapse_stop_condition.release()

                    if should_stop:
                        break
            finally:
                self._time_lapse_thread = None
                self.logger.info('Stopped time lapse recording')

        self._time_lapse_thread = threading.Thread(target=capture_thread)
        self._time_lapse_thread.start()

    @action
    def stop_time_lapse(self):
        """
        Stop a time lapse sequence if it's running
        """

        if not self._time_lapse_thread:
            self.logger.info('No time lapse thread is running')
            return

        self._time_lapse_stop_condition.acquire()
        self._time_lapse_stop_condition.notify_all()
        self._time_lapse_stop_condition.release()

        if self._time_lapse_thread:
            self._time_lapse_thread.join()

    # noinspection PyMethodOverriding
    @action
    def start_recording(self, video_file=None, directory=None, name_format='video_%04d.h264', duration=None,
                        split_duration=None, **opts):
        """
        Start recording to a video file or to multiple video files

        :param video_file: Path of the video file, if you want to keep the recording all in one file
        :type video_file: str

        :param directory: Path of the directory that will store the video files, if you want to split the recording
            on multiple files. Note that you need to specify either video_file (to save the recording to one single
            file) or directory (to split the recording on multiple files)
        :type directory: str

        :param name_format: If you're splitting the recording to multiple files, then you can specify the name format
            for those files (default: 'video_%04d.h264')
            on multiple files. Note that you need to specify either video_file (to save the recording to one single
            file) or directory (to split the recording on multiple files)
        :type name_format: str

        :param duration: Video duration in seconds (default: None, record until stop_recording is called)
        :type duration: float

        :param split_duration: If you're splitting the recording to multiple files, then you should specify how long
            each video should be in seconds
        :type split_duration: float

        :param opts: Extra options to pass to the camera (see
            https://www.raspberrypi.org/documentation/usage/camera/python/README.md)
        """

        if self._recording_thread:
            return None, 'A recording thread is already running'

        multifile = not video_file
        if multifile and not (directory and split_duration):
            return None, 'No video_file specified for single file capture and no directory/split_duration ' + \
                   'specified for multi-file split'

        camera = self._get_camera(**opts)
        video_file = os.path.abspath(os.path.expanduser(video_file))

        def recording_thread():
            try:
                if not multifile:
                    self.logger.info('Starting recording to video file {}'.format(video_file))
                    camera.start_recording(video_file, format='h264')
                    self._recording_stop_condition.acquire()
                    self._recording_stop_condition.wait(timeout=duration)
                    self._recording_stop_condition.release()
                    self.logger.info('Video recorded to {}'.format(video_file))
                    return

                self.logger.info('Starting recording video files to directory {}'.format(directory))
                i = 1
                end_time = None
                timeout = split_duration

                if duration is not None:
                    end_time = time.time() + duration
                    timeout = min(split_duration, duration)

                camera.start_recording(name_format % i, format='h264')
                self._recording_stop_condition.acquire()
                self._recording_stop_condition.wait(timeout=timeout)
                self._recording_stop_condition.release()
                self.logger.info('Video file {} saved'.format(name_format % i))

                while True:
                    i += 1
                    timeout = None

                    if end_time:
                        remaining_duration = end_time - time.time()
                        timeout = min(split_duration, remaining_duration)
                        if remaining_duration <= 0:
                            break

                    camera.split_recording(name_format % i)
                    self._recording_stop_condition.acquire()
                    should_stop = self._recording_stop_condition.wait(timeout=timeout)
                    self._recording_stop_condition.release()
                    self.logger.info('Video file {} saved'.format(name_format % i))

                    if should_stop:
                        break
            finally:
                try:
                    camera.stop_recording()
                except Exception as e:
                    self.logger.exception(e)

                self._recording_thread = None
                self.logger.info('Stopped camera recording')

        self._recording_thread = threading.Thread(target=recording_thread)
        self._recording_thread.start()

    @action
    def stop_recording(self, **kwargs):
        """
        Stop a camera recording
        """

        if not self._recording_thread:
            self.logger.info('No recording thread is running')
            return

        self._recording_stop_condition.acquire()
        self._recording_stop_condition.notify_all()
        self._recording_stop_condition.release()

        if self._recording_thread:
            self._recording_thread.join()

    # noinspection PyShadowingBuiltins
    @action
    def start_streaming(self, listen_port: Optional[int] = None, format='h264', **opts):
        """
        Start recording to a network stream

        :param listen_port: TCP listen port (default: `listen_port` configured value or 5000)
        :type listen_port: int

        :param format: Video stream format (default: h264)
        :type format: str

        :param opts: Extra options to pass to the camera (see
            https://www.raspberrypi.org/documentation/usage/camera/python/README.md)
        """

        if self._streaming_thread:
            return None, 'A streaming thread is already running'

        if not listen_port:
            listen_port = self.listen_port

        camera = self._get_camera(**opts)
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server_socket.bind(('0.0.0.0', listen_port))
        server_socket.listen(1)
        server_socket.settimeout(1)

        # noinspection PyBroadException
        def streaming_thread():
            try:
                self.logger.info('Starting streaming on port {}'.format(listen_port))

                while self._can_stream:
                    try:
                        sock = server_socket.accept()[0]
                        stream = sock.makefile('wb')
                        self.logger.info('Accepted client connection from {}'.format(sock.getpeername()))
                    except socket.timeout:
                        continue

                    try:
                        if stream:
                            camera.start_recording(stream, format=format)
                            while True:
                                camera.wait_recording(1)
                    except ConnectionError:
                        self.logger.info('Client closed connection')
                    finally:
                        if sock:
                            sock.close()
            finally:
                try:
                    server_socket.close()
                    camera.stop_recording()
                except:
                    pass

                try:
                    camera.close()
                except:
                    pass

                self._streaming_thread = None
                self.logger.info('Stopped camera stream')

        self._can_stream = True
        self._streaming_thread = threading.Thread(target=streaming_thread)
        self._streaming_thread.start()

    @action
    def stop_streaming(self):
        """
        Stop a camera streaming session
        """

        if not self._streaming_thread:
            self.logger.info('No recording thread is running')
            return

        self._can_stream = False

        if self._streaming_thread:
            self._streaming_thread.join()

    @action
    def is_streaming(self):
        """
        :return: True if the Pi Camera network streaming thread is running,
            False otherwise.
        """
        return self._streaming_thread is not None


# vim:sw=4:ts=4:et:



import time

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message.event.weather import NewWeatherConditionEvent, NewPrecipitationForecastEvent
from platypush.plugins.weather.buienradar import WeatherBuienradarPlugin


class WeatherBuienradarBackend(Backend):
    """
    Buienradar weather forecast backend. Listens for new weather or precipitation updates.

    Triggers:

        * :class:`platypush.message.event.weather.NewWeatherConditionEvent` when there is a weather condition update

    Requires:

        * The :mod:`platypush.plugins.weather.buienradar` plugin configured

    """

    def __init__(self, poll_seconds=300, **kwargs):
        super().__init__(**kwargs)
        self.poll_seconds = poll_seconds
        self.last_weather = None
        self.last_precip = None

    def run(self):
        super().run()
        plugin: WeatherBuienradarPlugin = get_plugin('weather.buienradar')
        self.logger.info('Initialized weather forecast backend')

        while not self.should_stop():
            weather = plugin.get_weather().output
            precip = plugin.get_precipitation().output
            del weather['measured']

            if precip != self.last_precip:
                self.bus.post(NewPrecipitationForecastEvent(average=precip.get('average'),
                                                            total=precip.get('total'),
                                                            time_frame=precip.get('time_frame')))

            if weather != self.last_weather:
                self.bus.post(NewWeatherConditionEvent(**weather))

            self.last_weather = weather
            self.last_precip = precip
            time.sleep(self.poll_seconds)


# vim:sw=4:ts=4:et:



import base64
import io
import os
import subprocess
import time

from platypush.plugins import Plugin, action


class CameraIrMlx90640Plugin(Plugin):
    """
    Plugin to interact with a `ML90640 <https://shop.pimoroni.com/products/mlx90640-thermal-camera-breakout>`_
    infrared thermal camera.

    In order to use this plugin you'll need to download and compile the
    `mlx90640 <https://github.com/pimoroni/mlx90640-library>`_ C++ bindings and examples for the device.
    Instructions on Raspbian::

        # Install the dependencies
        $ [sudo] apt-get install libi2c-dev
        $ cd $PLATYPUSH_SRC_DIR
        $ git submodule init
        $ git submodule update
        $ cd platypush/plugins/camera/ir/mlx90640/lib
        $ make clean
        $ make bcm2835
        $ make examples/rawrgb I2C_MODE=LINUX

    Requires:

        * **mlx90640-library** installation (see instructions above)
        * **PIL** image library (``pip install Pillow``)
    """

    _img_size = (32, 24)
    _rotate_values = {}

    def __init__(self, fps=16, skip_frames=2, scale_factor=1, rotate=0, rawrgb_path=None, **kwargs):
        """
        :param fps: Frames per seconds (default: 16)
        :param skip_frames: Number of frames to be skipped on sensor initialization/warmup (default: 2)
        :param scale_factor: The camera outputs 24x32 pixels artifacts. Use scale_factor to scale them up to a larger
            image (default: 1)
        :param rotate: Rotation angle in degrees (default: 0)
        :param rawrgb_path: Specify it if the rawrgb executable compiled from
            https://github.com/pimoroni/mlx90640-library is in another folder than
            `<directory of this file>/lib/examples`.
        """
        from PIL import Image
        super().__init__(**kwargs)

        self._rotate_values = {
            90: Image.ROTATE_90,
            180: Image.ROTATE_180,
            270: Image.ROTATE_270,
        }

        if not rawrgb_path:
            rawrgb_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'lib', 'examples', 'rawrgb')
        rawrgb_path = os.path.abspath(os.path.expanduser(rawrgb_path))

        assert fps > 0
        assert skip_frames >= 0
        assert os.path.isfile(rawrgb_path)

        self.fps = fps
        self.rotate = rotate
        self.skip_frames = skip_frames
        self.scale_factor = scale_factor
        self.rawrgb_path = rawrgb_path
        self._capture_proc = None

    def _is_capture_proc_running(self):
        return self._capture_proc is not None and self._capture_proc.poll() is None

    def _get_capture_proc(self, fps):
        if not self._is_capture_proc_running():
            fps = fps or self.fps
            self._capture_proc = subprocess.Popen([self.rawrgb_path, '{}'.format(fps)], stdin=subprocess.PIPE,
                                                  stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        return self._capture_proc

    # noinspection PyShadowingBuiltins
    @action
    def capture(self, output_file=None, frames=1, grayscale=False, fps=None, skip_frames=None, scale_factor=None,
                rotate=None, format='jpeg'):
        """
        Capture one or multiple frames and return them as raw RGB

        :param output_file: Can be either the path to a single image file or a format string
            (e.g. 'snapshots/image-{:04d}') in case of multiple frames. If not set the function will return a list of
            base64 encoded representations of the raw RGB frames, otherwise the list of captured files.
        :type output_file: str

        :param frames: Number of frames to be captured (default: 1). If None the capture process will proceed until
            `stop` is called.
        :type frames: int

        :param grayscale: Save the image as grayscale - black pixels will be colder, white pixels warmer
            (default: False)
        :type grayscale: bool

        :param fps: If set it overrides the fps parameter specified on the object (default: None)
        :type fps: int

        :param skip_frames: If set it overrides the skip_frames parameter specified on the object (default: None)
        :type skip_frames: int

        :param scale_factor: If set it overrides the scale_factor parameter specified on the object (default: None)
        :type scale_factor: float

        :param rotate: If set it overrides the rotate parameter specified on the object (default: None)
        :type rotate: int

        :param format: Output image format if output_file is not specified (default: jpeg).
            It can be jpg, png, gif or any format supported by PIL
        :type format: str

        :returns: list[str]. Each item is a base64 encoded representation of a frame in the specified format if
            output_file is not set, otherwise a list with the captured image files will be returned.
        """

        from PIL import Image
        fps = self.fps if fps is None else fps
        skip_frames = self.skip_frames if skip_frames is None else skip_frames
        scale_factor = self.scale_factor if scale_factor is None else scale_factor
        rotate = self._rotate_values.get(self.rotate if rotate is None else rotate, 0)

        size = self._img_size
        sleep_time = 1.0 / self.fps
        captured_frames = []
        n_captured_frames = 0
        files = set()
        camera = self._get_capture_proc(fps)

        while (frames is not None and n_captured_frames < frames) or (
                frames is None and self._is_capture_proc_running()):
            frame = camera.stdout.read(size[0] * size[1] * 3)

            if skip_frames > 0:
                time.sleep(sleep_time)
                skip_frames -= 1
                continue

            image = Image.frombytes('RGB', size, frame)

            if grayscale:
                image = self._convert_to_grayscale(image)
            if scale_factor != 1:
                size = tuple(i * scale_factor for i in size)
                image = image.resize(size, Image.ANTIALIAS)
            if rotate:
                image = image.transpose(rotate)

            if not output_file:
                temp = io.BytesIO()
                image.save(temp, format=format)
                frame = base64.encodebytes(temp.getvalue()).decode()
                captured_frames.append(frame)
            else:
                image_file = os.path.abspath(os.path.expanduser(output_file.format(n_captured_frames)))
                image.save(image_file)
                files.add(image_file)

            n_captured_frames += 1
            time.sleep(sleep_time)

        self.stop()
        return sorted([f for f in files]) if output_file else captured_frames

    @staticmethod
    def _convert_to_grayscale(image):
        from PIL import Image
        new_image = Image.new('L', image.size)

        for i in range(0, image.size[0]):
            for j in range(0, image.size[1]):
                r, g, b = image.getpixel((i, j))
                value = int(2.0 * r - 0.5 * g - 1.5 * b)

                if value > 255:
                    value = 255
                if value < 0:
                    value = 0

                new_image.putpixel((i, j), value)

        return new_image

    @action
    def stop(self):
        """
        Stop an ongoing capture session
        """
        if not self._is_capture_proc_running():
            return

        self._capture_proc.terminate()
        self._capture_proc.kill()
        self._capture_proc.wait()
        self._capture_proc = None

# vim:sw=4:ts=4:et:



"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

import threading
from typing import Any, Optional, Dict, Union

from platypush.plugins import Plugin, action


class GpioPlugin(Plugin):
    """
    Plugin to handle raw read/write operation on the Raspberry Pi GPIO pins.

    Requires:
        * **RPi.GPIO** (`pip install RPi.GPIO`)
    """

    def __init__(self, pins: Optional[Dict[str, int]] = None, mode: str = 'board', **kwargs):
        """
        :param mode: Specify 'board' if you want to use the board PIN numbers,
            'bcm' for Broadcom PIN numbers (default: 'board')
        :param pins: Configuration for the GPIO PINs as a name -> pin_number map.

        Example::

            {
                "LED_1": 14,
                "LED_2": 15,
                "MOTOR": 16,
                "SENSOR": 17
            }

        """

        super().__init__(**kwargs)
        self.mode = self._get_mode(mode)
        self._initialized = False
        self._init_lock = threading.RLock()
        self._initialized_pins = {}
        self.pins_by_name = pins if pins else {}
        self.pins_by_number = {number: name
                               for (name, number) in self.pins_by_name.items()}

    def _init_board(self):
        import RPi.GPIO as GPIO

        with self._init_lock:
            if self._initialized and GPIO.getmode():
                return

            GPIO.setmode(self.mode)
            self._initialized = True

    def _get_pin_number(self, pin):
        try:
            pin = int(str(pin))
        except ValueError:
            pin = self.pins_by_name.get(pin)
            if not pin:
                raise RuntimeError('Unknown PIN name: "{}"'.format(pin))

        return pin

    @staticmethod
    def _get_mode(mode_str: str) -> int:
        import RPi.GPIO as GPIO

        mode_str = mode_str.upper()
        assert mode_str in ['BOARD', 'BCM'], 'Invalid mode: {}'.format(mode_str)
        return getattr(GPIO, mode_str)

    @action
    def write(self, pin: Union[int, str], value: Union[int, bool],
              name: Optional[str] = None) -> Dict[str, Any]:
        """
        Write a byte value to a pin.

        :param pin: PIN number or configured name
        :param name: Optional name for the written value (e.g. "temperature" or "humidity")
        :param value: Value to write

        Response::

            output = {
                "name": <pin or metric name>,
                "pin": <pin>,
                "value": <value>,
                "method": "write"
            }
        """

        import RPi.GPIO as GPIO

        self._init_board()
        name = name or pin
        pin = self._get_pin_number(pin)

        if pin not in self._initialized_pins or self._initialized_pins[pin] != GPIO.OUT:
            GPIO.setup(pin, GPIO.OUT)
            self._initialized_pins[pin] = GPIO.OUT

        GPIO.setup(pin, GPIO.OUT)
        GPIO.output(pin, value)

        return {
            'name': name,
            'pin': pin,
            'value': value,
            'method': 'write',
        }

    @action
    def read(self, pin: Union[int, str], name: Optional[str] = None) -> Dict[str, Any]:
        """
        Reads a value from a PIN.

        :param pin: PIN number or configured name.
        :param name: Optional name for the read value (e.g. "temperature" or "humidity")

        Response::

            output = {
                "name": <pin number or pin/metric name>,
                "pin": <pin>,
                "value": <value>,
                "method": "read"
            }
        """

        import RPi.GPIO as GPIO

        self._init_board()
        name = name or pin
        pin = self._get_pin_number(pin)

        if pin not in self._initialized_pins:
            GPIO.setup(pin, GPIO.IN)
            self._initialized_pins[pin] = GPIO.IN

        val = GPIO.input(pin)

        return {
            'name': name,
            'pin': pin,
            'value': val,
            'method': 'read',
        }

    @action
    def get_measurement(self, pin=None):
        if pin is None:
            return self.read_all()
        return self.read(pin)

    @action
    def read_all(self):
        """
        Reads the values from all the configured PINs and returns them as a list. It will raise a RuntimeError if no
        PIN mappings were configured.
        """

        if not self.pins_by_number:
            raise RuntimeError("No PIN mappings were provided/configured")

        values = []
        for (pin, name) in self.pins_by_number.items():
            # noinspection PyUnresolvedReferences
            values.append(self.read(pin=pin, name=name).output)

        return values

    @action
    def cleanup(self):
        """
        Cleanup the state of the GPIO and resets PIN values.
        """
        import RPi.GPIO as GPIO

        with self._init_lock:
            if self._initialized:
                GPIO.cleanup()
                self._initialized_pins = {}
                self._initialized = False


# vim:sw=4:ts=4:et:



import os
import requests

from platypush.message import Message
from platypush.plugins import Plugin, action


class HttpRequestPlugin(Plugin):
    """
    Plugin for executing custom HTTP requests.

    Requires:

        * **requests** (``pip install requests``)

    Some example usages::

        # Execute a GET request on a JSON endpoint
        {
            "type": "request",
            "action": "http.request.get",
            "args": {
                "url": "http://remote-host/api/v1/entity",
                "params": {
                    "start": "2000-01-01"
                }
            }
        }

        # Execute an action on another Platypush host through HTTP interface
        {
            "type": "request",
            "action": "http.request.post",
            "args": {
                "url": "http://remote-host:8008/execute",
                "json": {
                    "type": "request",
                    "target": "remote-host",
                    "action": "music.mpd.play"
                }
            }
        }
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def _exec(method, url, output='text', **kwargs):
        """ Available output types: text (default), json, binary """

        if 'username' in kwargs and 'password' in kwargs:
            kwargs['auth'] = (kwargs.pop('username'), kwargs.pop('password'))

        method = getattr(requests, method)
        response = method(url, **kwargs)
        response.raise_for_status()

        if output == 'json':
            output = response.json()
        if output == 'binary':
            output = response.content
        else:
            output = response.text

        # noinspection PyBroadException
        try:
            # If the response is a Platypush JSON, extract it
            output = Message.build(output)
        except:
            pass

        return output

    @action
    def get(self, url, **kwargs):
        """
        Perform a GET request

        :param url: Target URL
        :type url: str

        :param kwargs: Additional arguments that will be transparently provided to the ``requests`` object, including
            but not limited to query params, data, JSON, headers etc.
            (see http://docs.python-requests.org/en/master/user/quickstart/#make-a-request)
        :type kwargs: dict
        """

        return self._exec(method='get', url=url, **kwargs)

    @action
    def post(self, url, **kwargs):
        """
        Perform a POST request

        :param url: Target URL
        :type url: str

        :param kwargs: Additional arguments that will be transparently provided to the ``requests`` object, including
            but not limited to query params, data, JSON, headers etc.
            (see http://docs.python-requests.org/en/master/user/quickstart/#make-a-request)
        :type kwargs: dict
        """

        return self._exec(method='post', url=url, **kwargs)

    @action
    def head(self, url, **kwargs):
        """
        Perform an HTTP HEAD request

        :param url: Target URL
        :type url: str

        :param kwargs: Additional arguments that will be transparently provided to the ``requests`` object, including
            but not limited to query params, data, JSON, headers etc.
            (see http://docs.python-requests.org/en/master/user/quickstart/#make-a-request)
        :type kwargs: dict
        """

        return self._exec(method='head', url=url, **kwargs)

    @action
    def put(self, url, **kwargs):
        """
        Perform a PUT request

        :param url: Target URL
        :type url: str

        :param kwargs: Additional arguments that will be transparently provided to the ``requests`` object, including
            but not limited to query params, data, JSON, headers etc.
            (see http://docs.python-requests.org/en/master/user/quickstart/#make-a-request)
        :type kwargs: dict
        """

        return self._exec(method='put', url=url, **kwargs)

    @action
    def delete(self, url, **kwargs):
        """
        Perform a DELETE request

        :param url: Target URL
        :type url: str

        :param kwargs: Additional arguments that will be transparently provided to the ``requests`` object, including
            but not limited to query params, data, JSON, headers etc.
            (see http://docs.python-requests.org/en/master/user/quickstart/#make-a-request)
        :type kwargs: dict
        """

        return self._exec(method='delete', url=url, **kwargs)

    @action
    def options(self, url, **kwargs):
        """
        Perform an HTTP OPTIONS request

        :param url: Target URL
        :type url: str

        :param kwargs: Additional arguments that will be transparently provided to the ``requests`` object, including
            but not limited to query params, data, JSON, headers etc.
            (see http://docs.python-requests.org/en/master/user/quickstart/#make-a-request)
        :type kwargs: dict
        """

        return self._exec(method='options', url=url, **kwargs)

    @action
    def download(self, url: str, path: str, **kwargs):
        """
        Locally download the content of a remote URL.

        :param url: URL to be downloaded.
        :param path: Path where the content will be downloaded on the local filesystem - must be a file name.
        """
        path = os.path.abspath(os.path.expanduser(path))
        content = self._exec(method='get', url=url, output='binary', **kwargs)

        with open(path, 'wb') as f:
            f.write(content)


# vim:sw=4:ts=4:et:
